"""
@file
@brief Implements simple stochastic gradient optimisation.
It is inspired from `_stochastic_optimizers.py
<https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/
neural_network/_stochastic_optimizers.py>`_.
"""
import numpy
from numpy.core._exceptions import UFuncTypeError


class BaseOptimizer:
    """
    Base stochastic gradient descent optimizer.

    :param coef: array, initial coefficient
    :param learning_rate_init: float
        The initial learning rate used. It controls the step-size
        in updating the weights.
    :param min_threshold: coefficients must be higher than *min_thresold*
    :param max_threshold: coefficients must be below than *max_thresold*

    The class holds the following attributes:

    * *learning_rate*: float, the current learning rate
    * *coef*: optimized coefficients
    * *min_threshold*, *max_threshold*: coefficients thresholds
    * *l2*: L2 regularization
    * *l1*: L1 regularization
    """

    def __init__(self, coef, learning_rate_init=0.1,
                 min_threshold=None, max_threshold=None,
                 l1=0., l2=0.):
        if not isinstance(coef, numpy.ndarray):
            raise TypeError("coef must be an array.")
        self.coef = coef
        self.learning_rate_init = learning_rate_init
        self.learning_rate = float(learning_rate_init)
        if (min_threshold is not None and
                not isinstance(min_threshold, (float, numpy.float64, numpy.float32))):
            raise TypeError("min_threshold must be a float")
        if (max_threshold is not None and
                not isinstance(max_threshold, (float, numpy.float64, numpy.float32))):
            raise TypeError("min_threshold must be a float")
        self.min_threshold = min_threshold
        self.max_threshold = max_threshold
        self.l1 = l1
        self.l2 = l2

    def _get_updates(self, grad):
        raise NotImplementedError("Must be overwritten.")  # pragma no cover

    def update_coef(self, grad):
        """
        Updates coefficients with given gradient.

        :param grad: array, gradient
        """
        if self.coef.shape != grad.shape:
            raise ValueError(
                "coef and grad must have the same shape coef {} != gradient {}."
                "".format(self.coef.shape, grad.shape))
        update = self._get_updates(grad)
        self.coef += update
        if self.min_threshold is not None:
            try:
                self.coef = numpy.maximum(self.coef, self.min_threshold)
            except UFuncTypeError:  # pragma: no cover
                raise RuntimeError(  # pylint: disable=W0707
                    "Unable to compute an upper bound with coef={} "
                    "max_threshold={}".format(self.coef, self.min_threshold))
        if self.max_threshold is not None:
            try:
                self.coef = numpy.minimum(self.coef, self.max_threshold)
            except UFuncTypeError:  # pragma: no cover
                raise RuntimeError(  # pylint: disable=W0707
                    "Unable to compute a lower bound with coef={} "
                    "max_threshold={}".format(self.coef, self.max_threshold))

    def iteration_ends(self, time_step):
        """
        Performs update to learning rate and potentially other states at the
        end of an iteration.
        """
        pass  # pragma: no cover

    def train(self, X, y, fct_loss, fct_grad, max_iter=100,
              early_th=None, verbose=False):
        """
        Optimizes the coefficients.

        :param X: datasets (array)
        :param y: expected target
        :param fct_loss: loss function, signature: `f(coef, X, y) -> float`
        :param fct_grad: gradient function,
            signature: `g(coef, x, y, i) -> array`
        :param max_iter: number maximum of iteration
        :param early_th: stops the training if the error goes below
            this threshold
        :param verbose: display information
        :return: loss

        The method keeps the best coefficients for the
        minimal loss.
        """
        if not isinstance(X, numpy.ndarray):
            raise TypeError("X must be an array.")
        if not isinstance(y, numpy.ndarray):
            raise TypeError("y must be an array.")
        if X.shape[0] != y.shape[0]:
            raise ValueError("X and y must have the same number of rows.")
        if any(numpy.isnan(X.ravel())):
            raise ValueError("X contains nan value.")
        if any(numpy.isnan(y.ravel())):
            raise ValueError("y contains nan value.")

        loss = fct_loss(self.coef, X, y)
        losses = [loss]
        best_coef = None
        best_loss = None
        if verbose:
            self._display_progress(0, max_iter, loss)
        n_samples = 0
        for it in range(max_iter):
            irows = numpy.random.choice(X.shape[0], X.shape[0])
            for irow in irows:
                grad = fct_grad(self.coef, X[irow, :], y[irow], irow)
                self._regularize_gradient(grad)
                if isinstance(verbose, int) and verbose >= 10:
                    self._display_progress(0, max_iter, loss, grad, 'grad')
                if numpy.isnan(grad).sum() > 0:
                    raise RuntimeError(  # pragma: no cover
                        "The gradient has nan values.")
                self.update_coef(grad)
                n_samples += 1

            self.iteration_ends(n_samples)
            loss = fct_loss(self.coef, X, y) + \
                self.loss_regularization(self.coef)
            if verbose:
                self._display_progress(it + 1, max_iter, loss)
            self.iter_ = it + 1
            losses.append(loss)
            if best_loss is None or loss < best_loss:
                best_loss = loss
                best_coef = self.coef.copy()
            if self._evaluate_early_stopping(
                    it, max_iter, losses, early_th, verbose=verbose):
                break
        self.coef = best_coef
        return best_loss

    def loss_regularization(self, coef):
        loss = 0
        if self.l1 > 0:
            loss += numpy.sum(numpy.abs(coef)) * self.l1
        if self.l2 > 0:
            loss += numpy.sum(coef ** 2) * self.l2
        return loss

    def _regularize_gradient(self, grad):
        """
        Applies regularization.
        """
        self.velocity_grad = grad
        if self.l2 > 0:
            grad += self.coef * self.l2
        if self.l1 > 0:
            grad += numpy.sign(self.coef) * self.l1

    def _evaluate_early_stopping(
            self,
            it,
            max_iter,
            losses,
            early_th,
            verbose=False):
        if len(losses) < 5 or early_th is None:
            return False
        if numpy.isnan(losses[-5]):
            if numpy.isnan(losses[-1]):
                if verbose:  # pragma: no cover
                    self._display_progress(it + 1, max_iter, losses[-1],
                                           losses=losses[-5:])
                return True
            return False
        if numpy.isnan(losses[-1]):
            if verbose:  # pragma: no cover
                self._display_progress(it + 1, max_iter, losses[-1],
                                       losses=losses[-5:])
            return True
        if abs(losses[-1] - losses[-5]) <= early_th:
            if verbose:  # pragma: no cover
                self._display_progress(it + 1, max_iter, losses[-1],
                                       losses=losses[-5:])
            return True
        return False

    def _display_progress(self, it, max_iter, loss, losses=None, msg=None):
        'Displays training progress.'
        mxc = numpy.abs(self.coef.ravel()).max()
        l1 = numpy.sum(numpy.abs(self.coef))
        l2 = numpy.sum(self.coef * self.coef)
        vl1 = numpy.sum(numpy.abs(self.velocity_grad))
        vl2 = numpy.sum(self.velocity_grad * self.velocity_grad)
        if losses is None:
            print('{}/{}: loss: {:1.4g} max(coef): {:1.2g} '
                  'l1={:1.2g}/{:1.2g} l2={:1.2g}/{:1.2g}'.format(
                      it, max_iter, loss, mxc, vl1, l1, vl2, l2))
        else:
            print('{}/{}: loss: {:1.4g} losses: {} max(coef): {:1.4g} '
                  'l1={:1.2g}/{:1.2g} l2={:1.2g}/{:1.2g}'.format(
                      it, max_iter, loss, losses, mxc, vl1, l1, vl2, l2))


class SGDOptimizer(BaseOptimizer):
    """
    Stochastic gradient descent optimizer with momentum.

    :param coef: array, initial coefficient
    :param learning_rate_init: float
        The initial learning rate used. It controls the step-size
        in updating the weights,
    :param lr_schedule: `{'constant', 'adaptive', 'invscaling'}`,
        learning rate schedule for weight updates,
        `'constant'` for a constant learning rate given by
        *learning_rate_init*. `'invscaling'` gradually decreases
        the learning rate *learning_rate_* at each time step *t*
        using an inverse scaling exponent of *power_t*.
        `learning_rate_ = learning_rate_init / pow(t, power_t)`,
        `'adaptive'`, keeps the learning rate constant to
        *learning_rate_init* as long as the training keeps decreasing.
        Each time 2 consecutive epochs fail to decrease the training loss by
        tol, or fail to increase validation score by tol if 'early_stopping'
        is on, the current learning rate is divided by 5.
    :param momentum: float
        Value of momentum used, must be larger than or equal to 0
    :param power_t: double
        The exponent for inverse scaling learning rate.
    :param early_th: stops if the error goes below that threshold
    :param min_threshold: lower bound for parameters (can be None)
    :param max_threshold: upper bound for parameters (can be None)
    :param l1: L1 regularization
    :param l2: L2 regularization

    The class holds the following attributes:

    * *learning_rate*: float, the current learning rate
    * velocity*: array, velocity that are used to update params

    .. exref::
        :title: Stochastic Gradient Descent applied to linear regression

        The following example how to optimize a simple linear regression.

        .. runpython::
            :showcode:

            import numpy
            from mlstatpy.optim import SGDOptimizer


            def fct_loss(c, X, y):
                return numpy.linalg.norm(X @ c - y) ** 2


            def fct_grad(c, x, y, i=0):
                return x * (x @ c - y) * 0.1


            coef = numpy.array([0.5, 0.6, -0.7])
            X = numpy.random.randn(10, 3)
            y = X @ coef

            sgd = SGDOptimizer(numpy.random.randn(3))
            sgd.train(X, y, fct_loss, fct_grad, max_iter=15, verbose=True)
            print('optimized coefficients:', sgd.coef)
    """

    def __init__(self, coef, learning_rate_init=0.1, lr_schedule='invscaling',
                 momentum=0.9, power_t=0.5, early_th=None,
                 min_threshold=None, max_threshold=None,
                 l1=0., l2=0.):
        super().__init__(coef, learning_rate_init,
                         min_threshold=min_threshold,
                         max_threshold=max_threshold,
                         l1=l1, l2=l2)
        self.lr_schedule = lr_schedule
        self.momentum = momentum
        self.power_t = power_t
        self.early_th = early_th
        self.velocity = numpy.zeros_like(coef)
        self.velocity_grad = numpy.zeros_like(coef)

    def iteration_ends(self, time_step):
        """
        Performs updates to learning rate and potential other states at the
        end of an iteration.

        :param time_step: int
            number of training samples trained on so far, used to update
            learning rate for 'invscaling'
        """
        if self.lr_schedule == 'invscaling':
            self.learning_rate = (float(self.learning_rate_init) /
                                  (time_step + 1) ** self.power_t)
        elif self.lr_schedule == 'constant':
            pass
        else:
            raise ValueError(  # pragma: no cover
                "Unexpected value: lr_schedule='{}'.".format(self.lr_schedule))

    def _get_updates(self, grad):
        """
        Gets the values used to update params with given gradients.

        :param grad: array, gradient
        :return: updates, array, the values to add to params
        """
        update = self.momentum * self.velocity - self.learning_rate * grad
        self.velocity = update
        return update

    def _display_progress(self, it, max_iter, loss, losses=None, msg='loss'):
        'Displays training progress.'
        mxc = numpy.abs(self.coef.ravel()).max()
        l1 = numpy.sum(numpy.abs(self.coef))
        l2 = numpy.sum(self.coef * self.coef)
        vl1 = numpy.sum(numpy.abs(self.velocity_grad))
        vl2 = numpy.sum(self.velocity_grad * self.velocity_grad)
        if losses is None:
            print('{}/{}: {}: {:1.4g} lr={:1.3g} max(coef): {:1.2g} '
                  'l1={:1.2g}/{:1.2g} l2={:1.2g}/{:1.2g}'.format(
                      it, max_iter, msg, loss, self.learning_rate, mxc,
                      vl1, l1, vl2, l2))
        else:
            print('{}/{}: {}: {:1.4g} lr={:1.3g} {}es: {} '  # pylint: disable=W1308
                  'max(coef): {:1.2g} l1={:1.2g}/{:1.2g} l2={:1.2g}/{:1.2g}'.format(  # pylint: disable=W1308
                      it, max_iter, msg, loss, self.learning_rate, msg, losses,
                      mxc, vl1, l1, vl2, l2))



# -*- coding: utf-8 -*-
"""
@file
@brief Algorithms about matrices.
"""
import warnings
import numpy
import numpy.linalg
from scipy.linalg.lapack import dtrtri  # pylint: disable=E0611


def gram_schmidt(mat, change=False):
    """
    Applies the `Gram–Schmidt process
    <https://en.wikipedia.org/wiki/Gram%E2%80%93Schmidt_process>`_.
    Due to performance, every row is considered as a vector.

    @param      mat         matrix
    @param      change      returns the matrix to change the basis
    @return                 new matrix or (new matrix, change matrix)

    The function assumes the matrix *mat* is
    horizontal: it has more columns than rows.

    .. note::
        The implementation could be improved
        by directly using :epkg:`BLAS` function.

    .. runpython::
        :showcode:

        import numpy
        from mlstatpy.ml.matrices import gram_schmidt

        X = numpy.array([[1., 2., 3., 4.],
                         [5., 6., 6., 6.],
                         [5., 6., 7., 8.]])
        T, P = gram_schmidt(X, change=True)
        print(T)
        print(P)
    """
    if len(mat.shape) != 2:
        raise ValueError("mat must be a matrix.")  # pragma: no cover
    if mat.shape[1] < mat.shape[0]:
        raise RuntimeError(  # pragma: no cover
            "The function only works if the number of rows is less "
            "than the number of columns.")
    if change:
        base = numpy.identity(mat.shape[0])
    # The following code is equivalent to:
    # res = numpy.empty(mat.shape)
    # for i in range(0, mat.shape[0]):
    #     res[i, :] = mat[i, :]
    #     for j in range(0, i):
    #         d = numpy.dot(res[j, :], mat[i, :])
    #         res[i, :] -= res[j, :] * d
    #         if change:
    #             base[i, :] -= base[j, :] * d
    #     d = numpy.dot(res[i, :], res[i, :])
    #     if d > 0:
    #         d **= 0.5
    #         res[i, :] /= d
    #         if change:
    #             base[i, :] /= d
    # But it is faster to write it this way:
    res = numpy.empty(mat.shape)
    for i in range(0, mat.shape[0]):
        res[i, :] = mat[i, :]
        if i > 0:
            d = numpy.dot(res[:i, :], mat[i, :])
            m = numpy.multiply(res[:i, :], d.reshape((len(d), 1)))
            m = numpy.sum(m, axis=0)
            res[i, :] -= m
            if change:
                m = numpy.multiply(base[:i, :], d.reshape((len(d), 1)))
                m = numpy.sum(m, axis=0)
                base[i, :] -= m
        d = numpy.dot(res[i, :], res[i, :])
        if d > 0:
            d **= 0.5
            res[i, :] /= d
            if change:
                base[i, :] /= d
    return (res, base) if change else res


def linear_regression(X, y, algo=None):
    """
    Solves the linear regression problem,
    find :math:`\\beta` which minimizes
    :math:`\\norme{y - X\\beta}`, based on the algorithm
    :ref:`Arbre de décision optimisé pour les régressions linéaires
    <algo_decision_tree_mselin>`.

    @param      X       features
    @param      y       targets
    @param      algo    None to use the standard algorithm
                        :math:`\\beta = (X'X)^{-1} X'y`,
                        `'gram'`, `'qr'`
    @return             beta

    .. runpython::
        :showcode:

        import numpy
        from mlstatpy.ml.matrices import linear_regression

        X = numpy.array([[1., 2., 3., 4.],
                         [5., 6., 6., 6.],
                         [5., 6., 7., 8.]]).T
        y = numpy.array([0.1, 0.2, 0.19, 0.29])
        beta = linear_regression(X, y, algo="gram")
        print(beta)

    ``algo=None`` computes :math:`\\beta = (X'X)^{-1} X'y`.
    ``algo='qr'`` uses a `QR <https://docs.scipy.org/doc/numpy/reference
    /generated/numpy.linalg.qr.html>`_ decomposition and calls function
    `dtrtri <https://docs.scipy.org/doc/scipy/reference/generated/scipy.
    linalg.lapack.dtrtri.html>`_ to invert an upper triangular matrix.
    ``algo='gram'`` uses :func:`gram_schmidt
    <mlstatpy.ml.matrices.gram_schmidt>` and then computes
    the solution of the linear regression (see above for a link
    to the algorithm).
    """
    if len(y.shape) != 1:
        warnings.warn(  # pragma: no cover
            "This function is not tested for a multidimensional linear regression.")
    if algo is None:
        inv = numpy.linalg.inv(X.T @ X)
        return inv @ (X.T @ y)
    if algo == "gram":
        T, P = gram_schmidt(X.T, change=True)
        # T = P X
        return (y.T @ T.T @ P).ravel()
    if algo == "qr":
        Q, R = numpy.linalg.qr(X, "full")
        Ri = dtrtri(R)[0]
        gamma = (y.T @ Q).ravel()
        return (gamma @ Ri.T).ravel()
    raise ValueError(  # pragma: no cover
        "Unknwown algo='{}'.".format(algo))


def norm2(X):
    """
    Computes the square norm for all rows of a
    matrix.
    """
    res = numpy.empty(X.shape[1])
    for i in range(X.shape[1]):
        res[i] = numpy.dot(X[i], X[i])
    return res


def streaming_gram_schmidt_update(Xk, Pk):
    """
    Updates matrix :math:`P_k` to produce :math:`P_{k+1}`
    which is the matrix *P* in algorithm
    :ref:`Streaming Linear Regression
    <algo_reg_lin_gram_schmidt_streaming>`.
    The function modifies the matrix *Pk*
    given as an input.

    @param      Xk      kth row
    @param      Pk      matrix *P* at iteration *k-1*
    """
    tki = Pk @ Xk
    idi = numpy.identity(Pk.shape[0])

    for i in range(0, Pk.shape[0]):
        val = tki[i]

        if i > 0:
            # for j in range(0, i):
            #    d = tki[j] * val
            #    tki[i] -= tki[j] * d
            #    Pk[i, :] -= Pk[j, :] * d
            #    idi[i, :] -= idi[j, :] * d

            dv = tki[:i] * val
            tki[i] -= numpy.dot(dv, tki[:i])
            dv = dv.reshape((i, 1))
            Pk[i, :] -= numpy.multiply(Pk[:i, :], dv).sum(axis=0)
            idi[i, :] -= numpy.multiply(idi[:i, :], dv).sum(axis=0)

        d = numpy.square(idi[i, :]).sum()  # pylint: disable=E1101
        d = tki[i] ** 2 + d
        if d > 0:
            d **= 0.5
            d = 1. / d
            tki[i] *= d
            Pk[i, :] *= d
            idi[i, :] *= d


def streaming_gram_schmidt(mat, start=None):
    """
    Solves the linear regression problem,
    find :math:`\\beta` which minimizes
    :math:`\\norme{y - X\\beta}`, based on
    algorithm :ref:`Streaming Gram-Schmidt
    <algo_reg_lin_gram_schmidt_streaming>`.

    @param      mat     matrix
    @param      start   first row to start iteration, ``X.shape[1]`` by default
    @return             iterator on

    The function assumes the matrix *mat* is
    horizontal: it has more columns than rows.

    .. runpython::
        :showcode:

        import numpy
        from mlstatpy.ml.matrices import streaming_gram_schmidt

        X = numpy.array([[1, 0.5, 10., 5., -2.],
                         [0, 0.4, 20, 4., 2.],
                         [0, 0.7, 20, 4., 2.]], dtype=float).T

        for i, p in enumerate(streaming_gram_schmidt(X.T)):
            print("iteration", i, "\\n", p)
            t = X[:i+3] @ p.T
            print(t.T @ t)
    """
    if len(mat.shape) != 2:
        raise ValueError("mat must be a matrix.")  # pragma: no cover
    if mat.shape[1] < mat.shape[0]:
        raise RuntimeError("The function only works if the number of rows is less "
                           "than the number of columns.")
    if start is None:
        start = mat.shape[0]
    mats = mat[:, :start]
    _, Pk = gram_schmidt(mats, change=True)
    yield Pk

    k = start
    while k < mat.shape[1]:
        streaming_gram_schmidt_update(mat[:, k], Pk)
        yield Pk
        k += 1


def streaming_linear_regression_update(Xk, yk, XkXk, bk):
    """
    Updates coefficients :math:`\\beta_k` to produce :math:`\\beta_{k+1}`
    in :ref:`l-piecewise-linear-regression`.
    The function modifies the matrix *Pk*
    given as an input.

    @param      Xk      kth row
    @param      yk      kth target
    @param      XkXk    matrix :math:`X_{1..k}'X_{1..k}', updated by the function
    @param      bk      current coefficient (updated by the function)
    """
    Xk = Xk.reshape((1, XkXk.shape[0]))
    xxk = Xk.T @ Xk
    XkXk += xxk
    err = Xk.T * (yk - Xk @ bk)
    bk[:] += (numpy.linalg.inv(XkXk) @ err).flatten()


def streaming_linear_regression(mat, y, start=None):
    """
    Streaming algorithm to solve a linear regression.
    See :ref:`l-piecewise-linear-regression`.

    @param      mat     features
    @param      y       expected target
    @return             iterator on coefficients

    .. runpython::
        :showcode:

        import numpy
        from mlstatpy.ml.matrices import streaming_linear_regression, linear_regression

        X  = numpy.array([[1, 0.5, 10., 5., -2.],
                          [0, 0.4, 20, 4., 2.],
                          [0, 0.7, 20, 4., 3.]], dtype=float).T
        y  = numpy.array([1., 0.3, 10, 5.1, -3.])

        for i, bk in enumerate(streaming_linear_regression(X, y)):
            bk0 = linear_regression(X[:i+3], y[:i+3])
            print("iteration", i, bk, bk0)
    """
    if len(mat.shape) != 2:
        raise ValueError("mat must be a matrix.")  # pragma: no cover
    if mat.shape[0] < mat.shape[1]:
        raise RuntimeError("The function only works if the number of rows is more "
                           "than the number of columns.")
    if len(y.shape) != 1:
        warnings.warn(  # pragma: no cover
            "This function is not tested for a multidimensional linear regression.")
    if start is None:
        start = mat.shape[1]

    Xk = mat[:start]
    XkXk = Xk.T @ Xk
    bk = numpy.linalg.inv(XkXk) @ (Xk.T @ y[:start])
    yield bk

    k = start
    while k < mat.shape[0]:
        streaming_linear_regression_update(mat[k], y[k], XkXk, bk)
        yield bk
        k += 1


def streaming_linear_regression_gram_schmidt_update(Xk, yk, Xkyk, Pk, bk):
    """
    Updates coefficients :math:`\\beta_k` to produce :math:`\\beta_{k+1}`
    in :ref:`Streaming Linear Regression
    <l-piecewise-linear-regression-gram_schmidt>`.
    The function modifies the matrix *Pk*
    given as an input.

    @param      Xk      kth row
    @param      yk      kth target
    @param      Xkyk    matrix :math:`X_{1..k}' y_{1..k}' (updated by the function)
    @param      Pk      Gram-Schmidt matrix produced by the streaming algorithm
                         (updated by the function)
    @param      bk      current coefficient (updated by the function)
    """
    Xk = Xk.T
    streaming_gram_schmidt_update(Xk, Pk)
    Xkyk += (Xk * yk).reshape(Xkyk.shape)
    bk[:] = Pk @ Xkyk @ Pk


def streaming_linear_regression_gram_schmidt(mat, y, start=None):
    """
    Streaming algorithm to solve a linear regression with
    Gram-Schmidt algorithm.
    See :ref:`l-piecewise-linear-regression-gram_schmidt`.

    @param      mat     features
    @param      y       expected target
    @return             iterator on coefficients

    .. runpython::
        :showcode:

        import numpy
        from mlstatpy.ml.matrices import streaming_linear_regression, linear_regression

        X  = numpy.array([[1, 0.5, 10., 5., -2.],
                          [0, 0.4, 20, 4., 2.],
                          [0, 0.7, 20, 4., 3.]], dtype=float).T
        y  = numpy.array([1., 0.3, 10, 5.1, -3.])

        for i, bk in enumerate(streaming_linear_regression(X, y)):
            bk0 = linear_regression(X[:i+3], y[:i+3])
            print("iteration", i, bk, bk0)
    """
    if len(mat.shape) != 2:
        raise ValueError("mat must be a matrix.")  # pragma: no cover
    if mat.shape[0] < mat.shape[1]:
        raise RuntimeError("The function only works if the number of rows is more "
                           "than the number of columns.")
    if len(y.shape) != 1:
        warnings.warn(  # pragma: no cover
            "This function is not tested for a multidimensional linear regression.")
    if start is None:
        start = mat.shape[1]

    Xk = mat[:start]
    xyk = Xk.T @ y[:start]
    _, Pk = gram_schmidt(Xk.T, change=True)
    bk = Pk @ xyk @ Pk
    yield bk

    k = start
    while k < mat.shape[0]:
        streaming_linear_regression_gram_schmidt_update(
            mat[k], y[k], xyk, Pk, bk)
        yield bk
        k += 1



"""
@file
@brief Helpers on logistic regression.
"""
import numpy
from pandas import DataFrame


def random_set_1d(n, kind):
    """
    Builds a random dataset as describes in example
    :ref:`l-example-logistic-decision`.

    @param      n       number of observations
    @param      kind    2, 3, 4 (see example)
    @return             array 2D
    """
    x = numpy.random.rand(n) * 3 - 1
    if kind == 3:
        y = numpy.empty(x.shape, dtype=numpy.int32)
        y[x < 0] = 0
        y[(x >= 0) & (x <= 1)] = 1
        y[x > 1] = 0
    elif kind == 2:
        y = numpy.empty(x.shape, dtype=numpy.int32)
        y[x < 0] = 0
        y[x >= 0] = 1
    elif kind == 4:
        y = numpy.empty(x.shape, dtype=numpy.int32)
        y[x < 0] = 0
        y[(x >= 0) & (x <= 0.8)] = 1
        y[(x >= 0.8) & (x <= 1.5)] = 0
        y[x > 1.5] = 1
    else:
        raise ValueError('kind must be in (2, 3, 4).')
    x2 = numpy.random.rand(n)
    return numpy.vstack([x, x2]).T, y


def plot_ds(X, y, ax=None, title=None):
    """
    Plots a dataset, *X* is a dataset with two
    features, *y* contains the binary labels.
    """
    if ax is None:
        import matplotlib.pyplot as plt  # pragma: no cover
        ax = plt.gca()  # pragma: no cover
    colors = {0: '#88CCCC', 1: '#CCCC88'}
    c = [colors[_] for _ in y]
    ax.scatter(X[:, 0], X[:, 1], c=c, s=20, edgecolor='k', lw=0.1)
    if title is not None:
        ax.set_title(title)
    return ax


def plog2(p):
    """
    Computes :math:`x \\ln_2 x`.
    """
    if p == 0:
        return 0
    return p * numpy.log(p) / numpy.log(2)


def logistic(x):
    """
    Computes :math:`\\frac{1}{1 + e^{-x}}`.
    """
    return 1. / (1. + numpy.exp(-x))


def likelihood(x, y, theta=1., th=0.):
    """
    Computes :math:`\\sum_i y_i f(\\theta (x_i - x_0)) + (1 - y_i) (1 - f(\\theta (x_i - x_0)))`
    where :math:`f(x_i)` is :math:`\\frac{1}{1 + e^{-x}}`.
    """
    lr = logistic((x - th) * theta)
    return y * lr + (1. - y) * (1 - lr)


def criteria(X, y):
    """
    Computes Gini, information gain, likelihood on a dataset
    with two features assuming the first coordinates is used to classify.

    @param      X       2D matrix
    @param      y       binary labels
    @return             dataframe
    """
    res = numpy.empty((X.shape[0], 8))
    res[:, 0] = X[:, 0]
    res[:, 1] = y
    order = numpy.argsort(res[:, 0])
    res = res[order, :].copy()
    x = res[:, 0].copy()
    y = res[:, 1].copy()

    for i in range(1, res.shape[0] - 1):
        # gini
        p1 = numpy.sum(y[:i]) / i
        p2 = numpy.sum(y[i:]) / (y.shape[0] - i)
        res[i, 2] = p1
        res[i, 3] = p2
        res[i, 4] = 1 - p1**2 - (1 - p1)**2 + 1 - p2**2 - (1 - p2)**2
        res[i, 5] = - plog2(p1) - plog2(1 - p1) - plog2(p2) - plog2(1 - p2)
        th = x[i]
        res[i, 6] = logistic(th)
        res[i, 7] = numpy.sum(likelihood(x, y, 1., th)) / res.shape[0]
    columns = ['X', 'y', 'p1', 'p2', 'Gini', 'Gain', 'lr', 'LL']
    return DataFrame(res[1:-1], columns=columns)


def criteria2(X, y):
    """
    Computes Gini, information gain, likelihood on a dataset
    with two features assuming the first coordinates is used to classify.

    @param      X       2D matrix
    @param      y       binary labels
    @return             dataframe
    """
    res = numpy.empty((X.shape[0], 5))
    res[:, 0] = X[:, 0]
    res[:, 1] = y
    order = numpy.argsort(res[:, 0])
    res = res[order, :].copy()
    x = res[:, 0].copy()
    y = res[:, 1].copy()

    for i in range(1, res.shape[0] - 1):
        # gini
        th = x[i]
        res[i, 2] = max(numpy.sum(likelihood(x, y, 1., th)),
                        numpy.sum(likelihood(x, y, -1., th))) / res.shape[0]
        res[i, 3] = max(numpy.sum(likelihood(x, y, 10., th)),
                        numpy.sum(likelihood(x, y, -10., th))) / res.shape[0]
        res[i, 4] = max(numpy.sum(likelihood(x, y, 100., th)),
                        numpy.sum(likelihood(x, y, -100., th))) / res.shape[0]
    columns = ['X', 'y', 'LL', 'LL-10', 'LL-100']
    return DataFrame(res[1:-1], columns=columns)



# -*- coding: utf-8 -*-
"""
@file
@brief Génère des images aléatoires.
"""
import math
import numpy
import numpy.random as nprnd  # pylint: disable=E1101


def random_noise_image(size, ratio=0.1):
    """
    Construit une image blanche de taille *size*,
    noircit aléatoirement *ratio x nb pixels*
    pixels.

    @param      size        taille de l'image
    @param      ratio       proportion de pixels à noircir
    @return                 :epkg:`numpy:array`
    """
    img = numpy.zeros((size[1], size[0]), dtype=numpy.float32)
    nb = int(ratio * size[0] * size[1])
    xr = nprnd.randint(0, size[0] - 1, nb)
    yr = nprnd.randint(0, size[1] - 1, nb)
    img[yr, xr] = 1
    return img


def random_segment_image(image, lmin=0.1, lmax=1., noise=0.01, density=1.):
    """
    Ajoute un segment aléatoire à une image.
    Génère des points le long d'un segment aléatoire.

    @param      image       :epkg:`numpy:array` (modifié par la fonction)
    @param      lmin        taille minimal du segment
    @param      lmax        taille maximam du segment
    @param      density     nombre de pixel à tirer le long de l'axe
    @param      noise       bruit
    @return                 dictionary with *size, angle, x1, y1, x2, y2, nbpoints*
    """
    def move_coordinate(x1, y1, x2, y2, X, Y):
        if x2 < 0:
            x1 -= x2
            x2 = 0
        x1 = min(max(x1, 0), X - 1)
        x2 = min(max(x2, 0), X - 1)
        y1 = min(max(y1, 0), Y - 1)
        y2 = min(max(y2, 0), Y - 1)
        size = int(((x1 - x2)**2 + (y1 - y2) ** 2) ** 0.5)
        return x1, y1, x2, y2, size

    mind = min(image.shape)
    lmin = int(mind * lmin)
    lmax = int(mind * lmax)
    size = nprnd.randint(lmin, lmax)
    angle = nprnd.random() * math.pi  # pylint: disable=E1101
    x1 = nprnd.randint(
        image.shape[1] - int(size * abs(math.cos(angle)) - 1))
    y1 = nprnd.randint(image.shape[0] - int(size * math.sin(angle) - 1))
    x2 = x1 + size * math.cos(angle)
    y2 = y1 + size * math.sin(angle)
    x1, y1, x2, y2, size = move_coordinate(
        x1, y1, x2, y2, image.shape[1], image.shape[0])
    t = nprnd.randint(0, size, int(size * density))
    xs = t * math.cos(angle) + x1
    ys = t * math.sin(angle) + x2
    noise = nprnd.randn(  # pylint: disable=E1101
        xs.shape[0] * 2).reshape(xs.shape[0], 2) * noise * mind
    xs += noise[:, 0]
    ys += noise[:, 1]
    xs = numpy.maximum(xs, numpy.zeros(xs.shape[0]))  # pylint: disable=E1111
    ys = numpy.maximum(ys, numpy.zeros(
        xs.shape[0]))  # pylint: disable=E1111,E1101,E1136
    xs = numpy.minimum(xs, numpy.zeros(  # pylint: disable=E1101,E1136
        xs.shape[0]) + image.shape[1] - 1)  # pylint: disable=E1111,E1101,E1136
    ys = numpy.minimum(ys, numpy.zeros(  # pylint: disable=E1101,E1136
        xs.shape[0]) + image.shape[0] - 1)  # pylint: disable=E1111,E1101,E1136
    xs = xs.astype(numpy.int32)  # pylint: disable=E1101
    ys = ys.astype(numpy.int32)  # pylint: disable=E1101
    image[ys, xs] = 1
    res = dict(size=size, angle=angle, x1=x1, y1=y1, x2=x2, y2=y2,
               nbpoints=xs.shape[0])  # pylint: disable=E1136
    return res



# -*- coding: utf-8 -*-
"""
@file
@brief Module *mlstatpy*.
Functions and examples associated to the
content of the :epkg:`mlstatpy`.
"""

__version__ = "0.3.593"
__author__ = "Xavier Dupré"
__github__ = "https://github.com/sdpython/mlstatpy"
__url__ = "http://www.xavierdupre.fr/app/mlstatpy/helpsphinx/index.html"
__license__ = "MIT License"
__blog__ = """
<?xml version="1.0" encoding="UTF-8"?>
<opml version="1.0">
    <head>
        <title>blog</title>
    </head>
    <body>
        <outline text="mlstatpy"
            title="mlstatpy"
            type="rss"
            xmlUrl="http://www.xavierdupre.fr/app/mlstatpy/helpsphinx/_downloads/rss.xml"
            htmlUrl="http://www.xavierdupre.fr/app/mlstatpy/helpsphinx/blog/main_0000.html" />
    </body>
</opml>
"""


def check(log=False):
    """
    Checks the library is working.
    It raises an exception.
    If you want to disable the logs:

    @param      log     if True, display information, otherwise
    @return             0 or exception
    """
    return True


def _setup_hook(use_print=False):
    """
    if this function is added to the module,
    the help automation and unit tests call it first before
    anything goes on as an initialization step.
    """
    # we can check many things, needed module
    # any others things before unit tests are started
    if use_print:
        print("Success: _setup_hook")



# -*- coding: utf-8 -*-
"""
@file
@brief About Voronoi Diagram
"""
import warnings
import numpy
from sklearn.linear_model import LinearRegression
from sklearn.metrics.pairwise import pairwise_distances
from mlinsights.mlmodel import QuantileLinearRegression


class VoronoiEstimationError(Exception):
    """
    Raised when the algorithm failed.
    """
    pass


def voronoi_estimation_from_lr(L, B, C=None, D=None, cl=0, qr=True, max_iter=None, verbose=False):
    """
    Determines a Voronoi diagram close to a convex
    partition defined by a logistic regression in *n* classes.
    :math:`M \\in \\mathbb{M}_{nd}` a row matrix :math:`(L_1, ..., L_n)`.
    Every border between two classes *i* and *j* is defined by:
    :math:`\\scal{L_i}{X} + B = \\scal{L_j}{X} + B`.

    The function looks for a set of points from which the Voronoi
    diagram can be inferred. It is done through a linear regression
    with norm *L1*. See :ref:`l-lrvor-connection`.

    @param      L           matrix
    @param      B           vector
    @param      C           additional conditions (see below)
    @param      D           addition condictions (see below)
    @param      cl          class on which the additional conditions applies
    @param      qr          use quantile regression
    @param      max_iter    number of condition to remove until convergence
    @param      verbose     display information while training
    @return                 matrix :math:`P \\in \\mathbb{M}_{nd}`

    The function solves the linear system:

    .. math::

        \\begin{array}{rcl}
        & \\Longrightarrow & \\left\\{\\begin{array}{l}\\scal{\\frac{L_i-L_j}{\\norm{L_i-L_j}}}{P_i + P_j} +
        2 \\frac{B_i - B_j}{\\norm{L_i-L_j}} = 0 \\\\
        \\scal{P_i-  P_j}{u_{ij}} - \\scal{P_i - P_j}{\\frac{L_i-L_j}{\\norm{L_i-L_j}}} \\scal{\\frac{L_i-L_j}{\\norm{L_i-L_j}}}{u_{ij}}=0
        \\end{array} \\right.
        \\end{array}

    If the number of dimension is big and
    the number of classes small, the system has
    multiple solution. Addition condition must be added
    such as :math:`CP_i=D` where *i=cl*, :math:`P_i`
    is the Voronoï point attached to class *cl*.
    `Quantile regression <https://fr.wikipedia.org/wiki/R%C3%A9gression_quantile>`_
    is not implemented in :epkg:`scikit-learn`.
    We use `QuantileLinearRegression <http://www.xavierdupre.fr/app/mlinsights/helpsphinx/mlinsights/mlmodel/quantile_regression.html>`_.

    After the first iteration, the function determines
    the furthest pair of points and removes it from the list
    of equations. If *max_iter* is None, the system goes until
    the number of equations is equal to the number of points * 2,
    otherwise it stops after *max_iter* removals. This is not the
    optimal pair to remove as they could still be neighbors but it
    should be a good heuristic.
    """
    labels_inv = {}
    nb_constraints = numpy.zeros((L.shape[0],))
    matL = []
    matB = []
    for i in range(0, L.shape[0]):
        for j in range(i + 1, L.shape[0]):
            li = L[i, :]
            lj = L[j, :]
            c = (li - lj)
            nc = (c.T @ c) ** 0.5

            # first condition
            mat = numpy.zeros((L.shape))
            mat[i, :] = c
            mat[j, :] = c
            d = -2 * (B[i] - B[j])
            matB.append(d)
            matL.append(mat.ravel())
            labels_inv[i, j, 'eq1'] = len(matL) - 1
            nb_constraints[i] += 1
            nb_constraints[j] += 1

            # condition 2 - hides multiple equation
            # we pick one
            coor = 0
            found = False
            while not found and coor < len(c):
                if c[coor] == 0:
                    coor += 1
                    continue
                if c[coor] == nc:
                    coor += 1
                    continue
                found = True
            if not found:
                raise ValueError(  # pragma: no cover
                    "Matrix L has two similar rows {0} and {1}. "
                    "Problem cannot be solved.".format(i, j))

            c /= nc
            c2 = c * c[coor]
            mat = numpy.zeros((L.shape))
            mat[i, :] = -c2
            mat[j, :] = c2

            mat[i, coor] += 1
            mat[j, coor] -= 1
            matB.append(0)
            matL.append(mat.ravel())
            labels_inv[i, j, 'eq2'] = len(matL) - 1
            nb_constraints[i] += 1
            nb_constraints[j] += 1

    nbeq = (L.shape[0] * (L.shape[0] - 1)) // 2
    matL = numpy.array(matL)
    matB = numpy.array(matB)

    if max_iter is None:
        max_iter = matL.shape[0] - matL.shape[1]

    if nbeq * 2 <= L.shape[0] * L.shape[1]:
        if C is None and D is None:
            warnings.warn(  # pragma: no cover
                "[voronoi_estimation_from_lr] Additional condition are required.")
        if C is not None and D is not None:
            matL = numpy.vstack([matL, numpy.zeros((1, matL.shape[1]))])
            a = cl * L.shape[1]
            b = a + L.shape[1]
            matL[-1, a:b] = C
            if not isinstance(D, float):
                raise TypeError("D must be a float not {0}".format(type(D)))
            matB = numpy.hstack([matB, [D]])
        elif C is None and D is None:
            pass  # pragma: no cover
        else:
            raise ValueError(
                "C and D must be None together or not None together.")

    sample_weight = numpy.ones((matL.shape[0],))
    tol = numpy.abs(matL.ravel()).max() * 1e-8 / matL.shape[0]
    order_removed = []
    removed = set()
    for it in range(0, max(max_iter, 1)):

        if qr:
            clr = QuantileLinearRegression(
                fit_intercept=False, max_iter=max(matL.shape))
        else:
            clr = LinearRegression(fit_intercept=False)

        clr.fit(matL, matB, sample_weight=sample_weight)
        score = clr.score(matL, matB, sample_weight)

        res = clr.coef_
        res = res.reshape(L.shape)

        # early stopping
        if score < tol:
            if verbose:
                print('[voronoi_estimation_from_lr] iter={0}/{1} score={2} tol={3}'.format(
                    it + 1, max_iter, score, tol))
            break

        # defines the best pair of points to remove
        dist2 = pairwise_distances(res, res)
        dist = [(d, n // dist2.shape[0], n % dist2.shape[1])
                for n, d in enumerate(dist2.ravel())]
        dist = [_ for _ in dist if _[1] < _[2]]
        dist.sort(reverse=True)

        # test equal points
        if dist[-1][0] < tol:
            _, i, j = dist[-1]
            eq1 = labels_inv[i, j, 'eq1']
            eq2 = labels_inv[i, j, 'eq2']
            if sample_weight[eq1] == 0 and sample_weight[eq2] == 0:
                sample_weight[eq1] = 1
                sample_weight[eq2] = 1
                nb_constraints[i] += 1
                nb_constraints[j] += 1
            else:
                keep = (i, j)
                pos = len(order_removed) - 1
                while pos >= 0:
                    i, j = order_removed[pos]
                    if i in keep or j in keep:
                        eq1 = labels_inv[i, j, 'eq1']
                        eq2 = labels_inv[i, j, 'eq2']
                        if sample_weight[eq1] == 0 and sample_weight[eq2] == 0:
                            sample_weight[eq1] = 1
                            sample_weight[eq2] = 1
                            nb_constraints[i] += 1
                            nb_constraints[j] += 1
                            break
                    pos -= 1
                if pos < 0:
                    raise VoronoiEstimationError(  # pragma: no cover
                        'Two classes have been merged in a single Voronoi point '
                        '(dist={0} < {1}). max_iter should be lower than '
                        '{2}'.format(dist[-1][0], tol, it))

        dmax, i, j = dist[0]
        pos = 0
        while (i, j) in removed or nb_constraints[i] == 0 or nb_constraints[j] == 0:
            pos += 1
            if pos == len(dist):
                break
            dmax, i, j = dist[pos]
        if pos == len(dist):
            break
        removed.add((i, j))
        order_removed.append((i, j))
        eq1 = labels_inv[i, j, 'eq1']
        eq2 = labels_inv[i, j, 'eq2']
        sample_weight[eq1] = 0
        sample_weight[eq2] = 0
        nb_constraints[i] -= 1
        nb_constraints[j] -= 1

        if verbose:
            print('[voronoi_estimation_from_lr] iter={0}/{1} score={2:.3g} tol={3:.3g} del P{4},{5} d={6:.3g}'.format(
                it + 1, max_iter, score, tol, i, j, dmax))

    return res



# -*- coding: utf-8 -*-
"""
@file
@brief Ce module inclut une classe qui permet de
parcourir tous les segments de l'image.
"""
import math
import copy
from .detection_segment_bord import SegmentBord_Commun
from .geometrie import Point


class SegmentBord(SegmentBord_Commun):
    """
    Définit un segment allant d'un bord à un autre de l'image,
    la classe va balayer toutes les orientations possibles,
    pour chaque orientation, elle va ensuite balayer toute l'image,

    * **angle** : orientation de balayage
    * **fin** : pour le balayage de l'image à une orientation donnee,
      le segment part d'un bord, fin désigne le dernier pixel
      du contour à envisager avec cette orientation
    * **vecteur** : vecteur directeur correspondant à angle
    * **bord1** : numero du bord (0,1,2,3) correspondant à ``self.a``,
      0 bord droit, 1 bord haut, 2 bord gauche, 3 bord bas
    * **dangle** : orientation à visiter 0, dangle, 2*dangle, 3*dangle, ...

    Pour parcourir les segments de l'image, on part d'un premier segment
    (methode premier), segment horizontal, bord 0,
    la méthode next passe au segment suivant jusqu'au dernier auquel
    cas la methode retourne *False*.
    Les segments sont orientés, si un gradient est proche du vecteur
    normal, l'opposé ne l'est pas.
    """

    # voir la remarque de la classe Point a propos de __slots__
    __slots__ = "angle", "fin", "vecteur", "bord1", "dangle"

    def __init__(self, dim, dangle=math.pi / 24.0):
        """initialise les dimensions et
        fait sorte que la classe contienne le premier segment"""
        SegmentBord_Commun.__init__(self, dim)
        self.premier()
        self.dangle = dangle

    def __str__(self):
        """permet d'afficher le segment"""
        s = SegmentBord_Commun.__str__(self)
        s += " -- bord " + str(self.bord1)
        s += " -- fin " + str(self.fin)
        s += " -- a " + "%3.1f" % (self.angle * 180 / math.pi)
        s += " -- vec " + "%2.2f,%2.2f" % (self.vecteur.x, self.vecteur.y)
        return s

    def premier(self):
        """définit le premier segment, horizontal, part du bord gauche"""
        self.angle = 0
        self.fin = Point(0, 0)
        self.calcul_vecteur()

    def milieu(self):
        """
        Un autre segment, pour débugger le programme,
        choisit une orientation pour laquelle on sait que le
        résultat doit être un segment significatif,
        la methode *next* ira plus vite au dernier segment.
        """
        self.angle = math.pi / 2
        self.calcul_vecteur()

    def calcul_vecteur(self):
        """
        En fonction de l'angle, calcule le vecteur direction du segment,
        ensuite fixe la première extrémité du segment ``self.a``
        et détermine la dernière des premières extremités pour
        un segment de cette orientation (angle).
        """
        self.vecteur = Point(math.cos(self.angle), math.sin(self.angle))

        # le vecteur est horizontal
        if abs(self.vecteur.x) < 1e-5:
            if self.vecteur.y > 0:
                # vers le haut, un seul bord pour la premiere extremite
                self.bord1 = 3
                self.a.x = 0
                self.a.y = 0
                self.fin.x = self.dim.x - 1
                self.fin.y = 0
                return self.calcul_vecteur_fin()
            else:
                # vers le bas, un seul bord pour la premiere extremite
                self.bord1 = 1
                self.a.x = self.dim.x - 1
                self.a.y = self.dim.y - 1
                self.fin.x = 0
                self.fin.y = self.dim.y - 1
                return self.calcul_vecteur_fin()

        # le vecteur est vertical
        if abs(self.vecteur.y) < 1e-5:
            if self.vecteur.x < 0:
                # vers la droite, un seul bord pour la premiere extremite
                self.bord1 = 0
                self.a.x = self.dim.x - 1
                self.a.y = 0
                self.fin.x = self.dim.x - 1
                self.fin.y = self.dim.y - 1
                return self.calcul_vecteur_fin()
            else:
                # vers la gauche, un seul bord pour la premiere extremite
                self.bord1 = 2
                self.a.x = 0
                self.a.y = self.dim.y - 1
                self.fin.x = 0
                self.fin.y = 0
                return self.calcul_vecteur_fin()

        if self.vecteur.x < 0:
            if self.vecteur.y < 0:
                # en bas a gauche, deux bords pour la premiere extremite
                self.bord1 = 0
                self.a.x = self.dim.x - 1
                self.a.y = 0
                self.fin.x = 0
                self.fin.y = self.dim.y - 1
                return self.calcul_vecteur_fin()
            else:
                # en haut a gauche, deux bords pour la premiere extremite
                self.bord1 = 3
                self.a.x = 0
                self.a.y = 0
                self.fin.x = self.dim.x - 1
                self.fin.y = self.dim.y - 1
                return self.calcul_vecteur_fin()
        else:
            if self.vecteur.y < 0:
                # en haut a droite, deux bords pour la premiere extremite
                self.bord1 = 1
                self.a.x = self.dim.x - 1
                self.a.y = self.dim.y - 1
                self.fin.x = 0
                self.fin.y = 0
                return self.calcul_vecteur_fin()
            else:
                # en bas a droite, deux bords pour la premiere extremite
                self.bord1 = 2
                self.a.x = 0
                self.a.y = self.dim.y - 1
                self.fin.x = self.dim.x - 1
                self.fin.y = 0
                return self.calcul_vecteur_fin()

    def calcul_vecteur_fin(self):
        """propose une seconde extrémité connaissant la première,
        beaucoup plus loin en conservant la meme orientation,
        du moment qu'on traverse l'image"""
        t = self.dim.x + self.dim.y
        self.b.x = int(self.a.x + self.vecteur.x * t)
        self.b.y = int(self.a.y + self.vecteur.y * t)

    def directeur(self):
        """retourne une copie du vecteur directeur"""
        return copy.copy(self.vecteur)

    def next(self):
        """passe au segment suivant dans le parcours de l'image"""

        if self.angle >= math.pi * 2 - 1e-5:
            # toute orientation visitee
            return False

        if self.a == self.fin:
            # tout vecteur visitee pour la meme orientation,
            # on passe a l'orientation suivante
            self.angle += self.dangle
            self.calcul_vecteur()
            if self.angle >= math.pi * 2:
                return False
            else:
                return True
        else:
            # on passe au segment suivant selon la meme orientation,
            # tout depend du bord sur lequel on est
            if self.bord1 == 0:
                # bord droit
                if self.a.y < self.dim.y - 1:
                    # pas besoin de changer de bord
                    self.a.y += 1
                else:
                    # on passe au bord suivant, bord haut
                    self.bord1 = 1
                    self.a.x -= 1
            elif self.bord1 == 1:
                # bord haut, meme raisonnement que pour le premier bord
                if self.a.x > 0:
                    self.a.x -= 1
                else:
                    self.bord1 = 2
                    self.a.y -= 1
            elif self.bord1 == 2:
                # bord gauche, meme raisonnement que pour le premier bord
                if self.a.y > 0:
                    self.a.y -= 1
                else:
                    self.bord1 = 3
                    self.a.x += 1
            elif self.bord1 == 3:
                # bord bas, meme raisonnement que pour le premier bord
                if self.a.x < self.dim.x - 1:
                    self.a.x += 1
                else:
                    self.bord1 = 0
                    self.a.y += 1
            # choisit une derniere extremite
            self.calcul_vecteur_fin()
            return True

    def calcul_bord2(self):
        """calcule précisément la second extrémité, parcourt la demi-droite
        jusqu'à sortir de l'image, le dernier point est la seconde extrémité"""
        a = self.a.arrondi()
        p = copy.copy(self.a)
        n = self.directeur()

        i = 0
        while a.x >= 0 and a.y >= 0 and a.x < self.dim.x and a.y < self.dim.y:
            p += n
            a = p.arrondi()
            i += 1

        self.b.x = a.x
        self.b.y = a.y
        r = -1
        if self.b.x < 0:
            self.b.x = 0
            r = 2
        if self.b.x >= self.dim.x:
            self.b.x = self.dim.x - 1
            r = 0
        if self.b.y < 0:
            self.b.y = 0
            r = 3
        if self.b.y >= self.dim.y:
            self.b.y = self.dim.y - 1
            r = 1
        return r



# -*- coding: utf-8 -*-
"""
@file
@brief Implements classic k-nn.
"""
import numpy
import numpy.linalg
from scipy.spatial.distance import euclidean


class NuagePoints:
    """
    Définit une classe de nuage de points.
    On suppose qu'ils sont définis par une matrice,
    chaque ligne est un élément.
    """

    def __init__(self):
        """
        constructeur
        """
        pass

    def fit(self, X, y=None):
        """
        Follows sklearn API.

        @param      X   training set
        @param      y   labels
        """
        self.nuage = X
        self.labels = y

    def kneighbors(self, X, n_neighbors=1, return_distance=True):
        """
        Return the k nearest neighbors.

        @param      X                   test set
        @param      n_neighbors         number of neighbors
        @param      return_distance     return distance as well
        @return                         array (dist), array (indices)
        """
        if n_neighbors != 1:
            raise NotImplementedError(  # pragma: no cover
                "Not implemented when n_neighbors != 1.")
        if not return_distance:
            raise NotImplementedError(  # pragma: no cover
                "Not implemented when return_distance is False.")

        dist = numpy.zeros(X.shape[0])
        ind = numpy.zeros(X.shape[0], dtype=numpy.int64)

        for i in range(X.shape[0]):
            row = X[i, :]
            row.resize((1, X.shape[1]))
            r = self.ppv(row)
            dist[i], ind[i] = r
        return dist, ind

    @property
    def shape(self):
        """
        Retourne la dimension du nuage.
        """
        return self.nuage.shape

    def distance(self, obj1, obj2):
        """
        Retourne une distance entre deux éléments.

        @param      obj1        object 1
        @param      obj2        object 2
        @return                 distance
        """
        return euclidean(obj1, obj2)

    def label(self, i):
        """
        Retourne le label de l'object d'indice ``i``.

        @param      i       indice
        @return             label or None if there is no label
        """
        return self.label[i] if self.label is not None else None

    def ppv(self, obj):
        """
        Retourne l'élément le plus proche de obj et sa distance avec obj.

        @param      obj     object
        @return             ``tuple(dist, index)``
        """
        ones = numpy.ones((self.nuage.shape[0], 1))
        mat = ones @ obj
        if len(mat.shape) == 1:
            mat.resize((mat.shape[0], 1))
        delta = self.nuage - mat
        norm = numpy.linalg.norm(delta, axis=1)
        i = numpy.argmin(norm)
        return norm[i], i



# -*- coding: utf-8 -*-
"""
@file
@brief Ce module determine si un segment est significatif, c'est à dire
si le nombre de fausses alarmes n'est pas trop élevé.
"""
from .geometrie import Segment


class SegmentNFA(Segment):
    """
    Un segment + un nombre de fausses alarmes,
    servira a memoriser les segments significatifs.
    """

    # voir la classe Point pour __slots__
    __slots__ = ("nfa",)

    def __init__(self, p1, p2, nfa):
        """segment + nombre de fausses alarmes"""
        Segment.__init__(self, p1, p2)
        self.nfa = nfa

    def __str__(self):
        """permet d'afficher ce segment"""
        s = Segment.__str__(self)
        s += "  nfa = " + str(self.nfa)
        return s

    def __lt__(self, o):
        return self.nfa < o.nfa


class InformationPoint:
    """
    Pour retenir toutes les informations relatives a un segment,
    une position (pos),
    la norme du gradient (norme),
    une information permettant de savoir si le gradient est
    proche du vecteur normal au segment (aligne)"""

    # voir la classe Point pour __slots__
    __slots__ = "pos", "aligne", "norme"

    def __init__(self, pos, aligne, norme):
        """constructeur, initialisation"""
        self.pos, self.aligne, self.norme = pos, aligne, norme

    def __str__(self):
        """permet d'afficher cette classe"""
        s = "aligne " + str(self.aligne)
        s += "  pix " + str(self.pos)
        s += "  gnor " + str(self.norme)
        return s


class LigneGradient:
    """
    Stocke toutes les informations relatives à un segment de l'image
    reliant deux points du contour, reçoit les informations
    de la methode @see me decoupe_gradient.

    A partir de là, un segment significatif a deux extrémités
    dont le gradient est dans le bon sens, on parcourt donc
    tous les couples d'extrémités possibles,
    d'abord la première (méthode @see me premier_chemin),
    puis les suivant (méthode @see me next_chemin)
    jusqu'au dernier couple.
    """

    def __init__(self, info_ligne, seuil_norme, seuil_nfa):
        """constructeur"""
        self.info_ligne = info_ligne  # informations
        self.nb = len(info_ligne)  # nombre de pixels
        self.seuil_norme = seuil_norme
        self.seuil_nfa = seuil_nfa

    def __len__(self):
        """
        Retourne le nombre de pixels dans le segment,
        peut etre different de la liste ``self.info_ligne``,
        ``self.nb`` est déterminé par @see me decoupe_gradient.
        """
        return self.nb

    def has_aligned_point(self):
        """
        Dit s'il existe des points alignés sur le segment.
        """
        return any(filter(lambda _: _.aligne, self.info_ligne))

    def extremite(self):
        """
        Comptabilise les indices des extremites possibles,
        les pixels choisis ont un gradient de la bonne orientation.
        """
        ext = []
        if self.has_aligned_point():
            for i in range(0, len(self)):
                if self.info_ligne[i].aligne and \
                        (i == 0 or i == len(self) - 1 or not self.info_ligne[i - 1].aligne or
                         not self.info_ligne[i + 1].aligne):
                    ext.append(i)
        return ext

    def premier_chemin(self, ext):
        """Retourne la premiere d'extremite possible."""
        return (0, 1)

    def next_chemin(self, ext, ij):
        """Retourne le couple suivant d'extrémités possibles,
        None, dans le cas contraire."""
        if ij[1] < len(ext) - 1:
            return (ij[0], ij[1] + 1)
        elif ij[0] < len(ext) - 2:
            return (ij[0] + 1, ij[0] + 2)
        else:
            return None

    def calcule_NFA(self, ext, ij, binomiale, nb_seg):
        """
        ``ext[ij[0]]``: premier indice du segment,
        ``ext[ij[1]]``: dernier indice du segment,
        calcule le nombre de NFA de ce segment
        (nombre de fausses alarmes).
        """
        ln = 0
        n = 0
        for i in range(ext[ij[0]], ext[ij[1]] + 1):
            if self.info_ligne[i].norme < self.seuil_norme:
                # on evite les petits gradients
                continue
            ln += 1
            if self.info_ligne[i].aligne:
                # on calcule un gradient dans le bon sens
                n += 1

        # on determine ensuite la probabilite d'un tel agencement
        # de fausses alarmes
        nfa = binomiale[(ln, n)]
        nfa *= nb_seg
        return nfa

    def segments_significatifs(self, binomiale, nb_seg):
        """
        Comptabilise le nombre de segments significatifs sur une ligne
        et les mémorise.
        """

        # on recense les extrémités possibles
        ext = self.extremite()

        if len(ext) < 2:
            # s'il n'y a qu'une extrémité possible,
            # ce n'est pas suffisant pour faire un segment
            return []

        # premier couple d'extrémités
        ij = self.premier_chemin(ext)
        res = []       # pour memoriser les segments significatifs

        while ij is not None:  # tant qu'il reste un couple d'extremite
            # probabilite de fausses alarmes pour ce segment
            nfa = self.calcule_NFA(ext, ij, binomiale, nb_seg)

            if nfa < self.seuil_nfa:
                # si cette proba est suffisamment faible,
                # l'agencement est un cas rare (non aleatoire),
                # il est significatif
                seg = SegmentNFA(self.info_ligne[ext[ij[0]]].pos,
                                 self.info_ligne[ext[ij[1]]].pos,
                                 nfa)
                # on l'ajoute a la liste
                res.append(seg)

            # on passe au segment suivant
            ij = self.next_chemin(ext, ij)

        # fin
        return res



# -*- coding: utf-8 -*-
"""
@file
@brief Implements optimized k-nn.
"""
import random
import numpy
from .kppv import NuagePoints


class NuagePointsLaesa (NuagePoints):
    """
    Implémente l'algorithme des plus proches voisins,
    version :ref:`LAESA <space_metric_algo_laesa_prime>`_
    """

    def __init__(self, nb_pivots):
        """
        Construit la classe

        @param      nb_pivots       number of pivots
        """
        NuagePoints.__init__(self)
        self.nb_pivots = nb_pivots

    def fit(self, X, y=None):
        """
        Follows sklearn API.

        @param      X   training set
        @param      y   labels
        """
        self.nuage = X
        self.labels = y
        self.selection_pivots(self.nb_pivots)

    def selection_pivots(self, nb):
        """
        Sélectionne *nb* pivots aléatoirements.

        @param      nb      nombre de pivots
        """
        nb = min(nb, self.nuage.shape[0])
        if nb == 1:
            self.pivots = [2]
        else:
            self.pivots = set()
            while len(self.pivots) < nb:
                i = random.randint(0, self.nuage.shape[0] - 1)
                if i not in self.pivots:
                    self.pivots.add(i)
            self.pivots = list(sorted(self.pivots))

        # on calcule aussi la distance de chaque éléments au pivots
        self.dist = numpy.zeros((self.nuage.shape[0], len(self.pivots)))
        for i in range(self.nuage.shape[0]):
            for j in range(len(self.pivots)):  # pylint: disable=C0200
                self.dist[i, j] = self.distance(
                    self.nuage[i, :], self.nuage[self.pivots[j], :])

    def ppv(self, obj):
        """
        Retourne l'élément le plus proche de obj et sa distance avec obj,
        utilise la sélection à l'aide pivots

        @param      obj     object
        @return             ``tuple(distance, index)``
        """

        # initialisation
        dp = [(self.distance(obj, self.nuage[p, :]), p, i)
              for i, p in enumerate(self.pivots)]

        # pivots le plus proche
        dm, im, _ = min(dp)

        # améliorations
        for i in range(0, self.nuage.shape[0]):

            # on regarde si un pivot permet d'éliminer l'élément i
            calcul = True
            for d, p, ip in dp:
                delta = abs(d - self.dist[i, ip])
                if delta > dm:
                    calcul = False
                    break

            # dans le cas contraire on calcule la distance
            if calcul:
                d = self.distance(obj, self.nuage[i, :])
                if d < dm:
                    dm = d
                    im = i

        return dm, im



# -*- coding: utf-8 -*-
"""
@file
@brief First approach for a edit distance between two graphs.

See :ref:`l-graph_distance`.
"""
import copy
import re


class _Vertex:

    __slots__ = ('nb', 'label', 'weight')

    def __init__(self, nb, label, weight):
        self.nb = nb         # kind of id
        self.label = label      # label
        self.weight = weight


class Vertex(_Vertex):
    """
    Defines a vertex of a graph.
    """

    def __init__(self, nb, label, weight):
        """
        constructor
        @param      nb      (int) index of the vertex
        @param      label   (str) label
        @para       weight  (float)
        """
        _Vertex.__init__(self, nb, label, weight)
        self.pair = (None, None)
        self.edges = {}
        self.predE = {}
        self.succE = {}

    def __str__(self):
        """
        usual
        """
        return '{}'.format(self.Label)

    def __repr__(self):
        """
        usual
        """
        return "Vertex({}, {}, {})".format(repr(self.nb), repr(self.Label), self.weight)

    def is_vertex(self):
        """
        returns True
        """
        return True

    def is_edge(self):
        """
        returns False
        """
        return False

    @property
    def Label(self):
        """
        returns the label
        """
        return self.label


class _Edge:

    __slots__ = ('from_', 'to', 'label', 'weight', 'nb')

    def __init__(self, from_, to, label, weight):
        self.from_, self.to = from_, to
        self.nb = from_, to  # pylint: disable=E1101
        self.label = label


class Edge(_Edge):
    """
    Defines an edge of a graph.
    """

    def __init__(self, from_, to, label, weight):
        """
        @param  from_       (int)
        @param  to          (int)
        @param  label       (str)
        @param  weight      (float)

        ``'00'`` means the beginning of a graph, ``'11'`` the end.
        """
        _Edge.__init__(self, from_, to, label, weight)
        self.pair = (None, None)
        self.weight = weight
        if self.from_ == "00" and self.to == "00":
            raise AssertionError("should not happen")  # pragma: no cover
        if self.from_ == "11" and self.to == "11":
            raise AssertionError("should not happen")  # pragma: no cover

    def __str__(self):
        """
        usual
        """
        return "{} -> {} [{}]".format(
            self.nb[0], self.nb[1], self.Label)  # pylint: disable=E1101

    def __repr__(self):
        """
        usual
        """
        return "Edge({}, {}, {}, {})".format(
            repr(self.nb[0]), repr(self.nb[1]),  # pylint: disable=E1101
            repr(self.Label), self.weight)  # pylint: disable=E1101

    def is_vertex(self):
        """
        returns False
        """
        return False

    def is_edge(self):
        """
        returns True
        """
        return True

    @property
    def Label(self):
        """
        returns the label
        """
        return self.label


class GraphDistance:
    """
    Defines a graph to compute a distance between two graphs.

    .. exref::
        :title: Compute a distance between two graphs.

        See :ref:`l-graph_distance`.

        .. runpython::
            :showcode:

            import copy
            from mlstatpy.graph import GraphDistance

            # We define two graphs as list of edges.
            graph1 = [("a", "b"), ("b", "c"), ("b", "X"), ("X", "c"),
                      ("c", "d"), ("d", "e"), ("0", "b")]
            graph2 = [("a", "b"), ("b", "c"), ("b", "X"), ("X", "c"),
                      ("c", "t"), ("t", "d"), ("d", "e"), ("d", "g")]

            # We convert them into objects GraphDistance.
            graph1 = GraphDistance(graph1)
            graph2 = GraphDistance(graph2)

            distance, graph = graph1.distance_matching_graphs_paths(graph2, use_min=False)

            print("distance", distance)
            print("common paths:", graph)
    """

    # graph is a dictionary
    @staticmethod
    def get_list_of_vertices(graph):
        edges = [edge[:2] for edge in graph]
        unique = {}
        for i, j in edges:
            unique[i] = unique[j] = 1
        vertices = list(unique.keys())
        vertices.sort()
        return vertices

    def __init__(self, edge_list, vertex_label=None, add_loop=False,
                 weight_vertex=1., weight_edge=1.):
        """
        constructor

        @param      edge_list        list of edges
        @param      add_loop         automatically add a loop on each vertex (an edge from a vertex to itself)
        @param      weight_vertex    weight for every vertex
        @param      weight_edge      weight for every edge
        """
        if vertex_label is None:
            vertex_label = dict()
        if isinstance(edge_list, str):
            self.load_from_file(edge_list, add_loop)
        else:
            self.vertices = {}
            self.edges = {}
            self.labelBegin = "00"
            self.labelEnd = "11"
            vid = GraphDistance.get_list_of_vertices(edge_list)
            for u in vid:
                self.vertices[u] = Vertex(
                    u, vertex_label.get(u, str(u)), weight_vertex)
            for e in edge_list:
                i, j = e[:2]
                ls = "" if len(e) < 3 else e[2]
                self.edges[i, j] = Edge(i, j, str(ls), weight_edge)
            self._private__init__(add_loop, weight_vertex, weight_edge)

    def __getitem__(self, index):
        """
        returns a vertex or an edge if no vertex with the given index was found
        @param      index   id (index) to look for
        @return             Vertex or Edge
        """
        if isinstance(index, str):
            return self.vertices[index]
        if isinstance(index, tuple):
            return self.edges[index]
        raise KeyError(  # pragma: no cover
            "unable to get element " + str(index))

    @staticmethod
    def load_from_file(filename, add_loop):
        """
        loads a graph from a file
        @param      filename        file name
        @param      add_loop         @see me __init__
        """
        lines = open(filename, "r").readlines()  # pylint: disable=R1732
        regV = re.compile("\\\"?([a-z0-9_]+)\\\"? *[[]label=\\\"(.*)\\\"[]]")
        regE = re.compile("\\\"?([a-z0-9_]+)\\\"? *-> *\\\"?" +
                          "([a-z0-9_]+)\\\"? *[[]label=\\\"(.*)\\\"[]]")
        edge_list = []
        vertex_label = {}
        for line in lines:
            line = line.strip("\r\n ;")
            ed = regE.search(line)
            ve = regV.search(line)
            if ed:
                g = ed.groups()
                edge_list.append((g[0], g[1], g[2]))
            elif ve:
                g = ve.groups()
                vertex_label[g[0]] = g[1]
        if len(vertex_label) == 0 or len(edge_list) == 0:
            raise OSError(  # pragma: no cover
                "Unable to parse file %r." % filename)
        return GraphDistance(edge_list, vertex_label, add_loop)

    def _private__init__(self, add_loop, weight_vertex, weight_edge):
        if add_loop:
            for k in self.vertices:
                if k not in (self.labelBegin, self.labelEnd):
                    self.edges[k, k] = Edge(k, k, "", weight_edge)
        self.connect_root_and_leave(weight_vertex, weight_edge)
        self.compute_predecessor()
        self.compute_successor()

    def connect_root_and_leave(self, weight_vertex, weight_edge):
        order = self.get_order_vertices()
        roots = [v for v, k in order.items() if k == 0]
        vert = {}
        for o in order:
            vert[o] = 0
        for k in self.edges:
            if k[0] != k[1]:
                vert[k[0]] += 1
        for r in roots:
            if self.labelBegin not in self.vertices:
                self.vertices[self.labelBegin] = Vertex(
                    self.labelBegin, self.labelBegin, weight_vertex)
            if r != self.labelBegin:
                self.edges[self.labelBegin, r] = Edge(
                    self.labelBegin, r, "", weight_edge)

        leaves = [k for k, v in vert.items() if v == 0]
        for r in leaves:
            if self.labelEnd not in self.vertices:
                self.vertices[self.labelEnd] = Vertex(
                    self.labelEnd, self.labelEnd, weight_vertex)
            if r != self.labelEnd:
                self.edges[r, self.labelEnd] = Edge(
                    r, self.labelEnd, "", weight_edge)

    def get_order_vertices(self):
        edges = self.edges
        order = {}
        for k in edges:
            order[k[0]] = 0
            order[k[1]] = 0

        modif = 1
        while modif > 0:
            modif = 0
            for k in edges:
                i, j = k
                if i != j and order[j] <= order[i]:
                    order[j] = order[i] + 1
                    modif += 1

        return order

    def __str__(self):
        """
        usual
        """
        li = []
        for v in self.vertices.values():
            li.append(str(v))
        for k, e in self.edges.items():
            li.append(str(e))
        return "\n".join(li)

    def __repr__(self):
        """
        usual
        """
        edges = ", ".join(repr(v) for _, v in sorted(self.edges.items()))
        vertices = ", ".join("'{}': {}".format(k, repr(v))
                             for k, v in sorted(self.vertices.items()))
        return "GraphDistance(\n    [{0}],\n    {{{1}}})".format(edges, vertices)

    def compute_predecessor(self):
        """
        usual
        """
        pred = {}
        for i, j in self.edges:
            if j not in pred:
                pred[j] = {}
            pred[j][i, j] = 0
        for k, v in pred.items():
            for n in v:
                self.vertices[k].predE[n] = self.edges[n]

    def compute_successor(self):
        succ = {}
        for i, j in self.edges:
            if i not in succ:
                succ[i] = {}
            succ[i][i, j] = i, j
        for k, v in succ.items():
            for n in v:
                self.vertices[k].succE[n] = self.edges[n]

    def get_matching_functions(self, function_mach_vertices, function_match_edges,
                               cost=False):
        """
        returns default matching functions between two vertices and two edges
        @param      function_mach_vertices   if not None, this function is returned, othewise, it returns a new fonction.
                                             See below.
        @param      function_match_edges     if not None, this function is returned, othewise, it returns a new fonction.
                                             See below.
        @param      cost                     if True, the returned function should return a float, otherwise a boolean
        @return                              a pair of functions

        Example for * if cost is False:

        ::

            lambda v1,v2,g1,g2,w1,w2 : v1.label == v2.label

        Example for *function_mach_vertices* if cost is True:

        ::

            def tempF1 (v1,v2,g1,g2,w1,w2) :
                if v1 is not None and not v1.is_vertex() : raise TypeError("should be a vertex")
                if v2 is not None and not v2.is_vertex() : raise TypeError("should be a vertex")
                if v1 is None and v2 is None : return 0
                elif v1 is None or v2 is None :
                    return v2.weight*w2 if v1 is None else v1.weight*w1
                else :
                    return 0 if v1.label == v2.label else 0.5*(v1.weight*w1 + v2.weight*w2)

        Example for *function_match_edges* if cost is False:

        ::

            lambda e1,e2,g1,g2,w1,w2 : e1.label == e2.label and
                        (e1.from_ != e1.to or e2.from_ != e2.to) and
                        (e1.from_ != self.labelBegin or e1.to != self.labelBegin) and
                        (e1.from_ != self.labelEnd or e1.to != self.labelEnd)

        Example if cost is True:

        ::

            def tempF2 (e1,e2,g1,g2,w1,w2) :
                if e1 is not None and not e1.is_edge() : raise TypeError("should be an edge")
                if e2 is not None and not e2.is_edge() : raise TypeError("should be an edge")
                if e1 is None and e2 is None : return 0
                elif e1 is None or e2 is None :
                    return e2.weight*w2 if e1 is None else e1.weight*w1
                elif e1.label != e2.label : return 0.5*(e1.weight*w1 + e2.weight*w2)
                else :
                    lab1 = g1.vertices [e1.from_].label == g2.vertices [e2.from_].label
                    lab2 = g1.vertices [e1.to].label == g2.vertices [e2.to].label
                    if lab1 and lab2 : return 0
                    else :  return e1.weight*w1 + e2.weight*w2

        """
        if cost:

            if function_mach_vertices is None:
                def tempF1_vertex(v1, v2, g1, g2, w1, w2):
                    if v1 is None:
                        if v2 is None:
                            return 0.
                        if not v2.is_vertex():
                            raise TypeError(  # pragma: no cover
                                "v2 should be a vertex")
                        return v2.weight * w2
                    elif v2 is None:
                        if not v1.is_vertex():
                            raise TypeError(  # pragma: no cover
                                "v1 should be a vertex")
                        if not v1.is_vertex():
                            raise TypeError(  # pragma: no cover
                                "v1 should be a vertex")
                        return v1.weight * w1
                    else:
                        if not v1.is_vertex():
                            raise TypeError(  # pragma: no cover
                                "v1 should be a vertex")
                        if not v2.is_vertex():
                            raise TypeError(  # pragma: no cover
                                "v2 should be a vertex")
                        return (
                            0 if v1.label == v2.label
                            else 0.5 * (v1.weight * w1 + v2.weight * w2))

                function_mach_vertices = tempF1_vertex

            if function_match_edges is None:
                def tempF2_edge(e1, e2, g1, g2, w1, w2):
                    if e1 is not None and not e1.is_edge():
                        raise TypeError("should be an edge")
                    if e2 is not None and not e2.is_edge():
                        raise TypeError("should be an edge")
                    if e1 is None and e2 is None:
                        return 0
                    elif e1 is None or e2 is None:
                        return e2.weight * w2 if e1 is None else e1.weight * w1
                    elif e1.label != e2.label:
                        return 0.5 * (e1.weight * w1 + e2.weight * w2)
                    else:
                        lab1 = g1.vertices[e1.from_].label == g2.vertices[e2.from_].label
                        lab2 = g1.vertices[e1.to].label == g2.vertices[e2.to].label
                        if lab1 and lab2:
                            return 0
                        else:
                            return e1.weight * w1 + e2.weight * w2

                function_match_edges = tempF2_edge
        else:
            if function_mach_vertices is None:
                function_mach_vertices = \
                    lambda v1, v2, g1, g2, w1, w2: \
                    v1.label == v2.label
            if function_match_edges is None:
                function_match_edges = \
                    lambda e1, e2, g1, g2, w1, w2: \
                    e1.label == e2.label and \
                    (e1.from_ != e1.to or e2.from_ != e2.to) and \
                    (e1.from_ != self.labelBegin or e1.to != self.labelBegin) and \
                    (e1.from_ != self.labelEnd or e1.to != self.labelEnd)
        return function_mach_vertices, function_match_edges

    def common_paths(self, graph2,
                     function_mach_vertices=None,
                     function_match_edges=None,
                     noClean=False):
        function_mach_vertices, function_match_edges = \
            self.get_matching_functions(
                function_mach_vertices, function_match_edges)
        g = GraphDistance([])
        vfirst = Vertex(self.labelBegin, "%s-%s" % (self.labelBegin, self.labelBegin),
                        (self.vertices[self.labelBegin].weight +
                         graph2.vertices[self.labelBegin].weight) / 2)
        g.vertices[self.labelBegin] = vfirst
        vfirst.pair = self.vertices[
            self.labelBegin], graph2.vertices[self.labelBegin]

        modif = 1
        while modif > 0:
            modif = 0
            add = {}
            for k, v in g.vertices.items():

                v1, v2 = v.pair
                if len(v.succE) == 0:
                    for e1 in v1.succE:
                        for e2 in v2.succE:
                            oe1 = self.edges[e1]
                            oe2 = graph2.edges[e2]
                            if function_match_edges(oe1, oe2, self, graph2, 1., 1.):
                                tv1 = self.vertices[oe1.to]
                                tv2 = graph2.vertices[oe2.to]
                                if function_mach_vertices(tv1, tv2, self, graph2, 1., 1.):
                                    # we have a match
                                    ii = "%s-%s" % (tv1.nb, tv2.nb)
                                    if tv1.nb == self.labelEnd and tv2.nb == self.labelEnd:
                                        ii = self.labelEnd
                                    lab = "%s-%s" % (tv1.label, tv2.label) \
                                        if tv1.label != tv2.label else tv1.label
                                    tv = Vertex(
                                        ii, lab, (tv1.weight + tv2.weight) / 2)
                                    lab = "%s-%s" % (oe1.label, oe2.label) \
                                        if oe1.label != oe2.label else oe1.label
                                    ne = Edge(v.nb, tv.nb, lab,
                                              (oe1.weight + oe2.weight) / 2)
                                    add[tv.nb] = tv
                                    g.edges[ne.from_, ne.to] = ne
                                    ne.pair = oe1, oe2
                                    tv.pair = tv1, tv2
                                    v.succE[ne.from_, ne.to] = ne
                                    modif += 1
            for k, v in add.items():
                g.vertices[k] = v

        if not noClean:
            # g.connect_root_and_leave()
            g.compute_predecessor()
            g.clean_dead_ends()
        return g

    def clean_dead_ends(self):
        edgesToKeep = {}
        verticesToKeep = {}
        if self.labelEnd in self.vertices:
            v = self.vertices[self.labelEnd]
            verticesToKeep[v.nb] = False

            modif = 1
            while modif > 0:
                modif = 0
                add = {}
                for k, v in verticesToKeep.items():
                    if v:
                        continue
                    modif += 1
                    verticesToKeep[k] = True
                    for pred, vv in self.vertices[k].predE.items():
                        edgesToKeep[pred] = True
                        add[vv.from_] = verticesToKeep.get(vv.from_, False)
                for k, v in add.items():
                    verticesToKeep[k] = v

            remove = {}
            for k in self.vertices:
                if k not in verticesToKeep:
                    remove[k] = True
            for k in remove:
                del self.vertices[k]

            remove = {}
            for k in self.edges:
                if k not in edgesToKeep:
                    remove[k] = True
            for k in remove:
                del self.edges[k]
        else:
            self.vertices = {}
            self.edges = {}

    def enumerate_all_paths(self, edges_and_vertices, begin=None):
        if begin is None:
            begin = []
        if len(self.vertices) > 0 and len(self.edges) > 0:
            if edges_and_vertices:
                last = begin[-1] if len(begin) > 0 \
                    else self.vertices[self.labelBegin]
            else:
                last = self.vertices[begin[-1].to] if len(begin) > 0 \
                    else self.vertices[self.labelBegin]

            if edges_and_vertices and len(begin) == 0:
                begin = [last]

            for ef in last.succE:
                e = self.edges[ef]
                path = copy.copy(begin)
                v = self.vertices[e.to]
                if e.to == e.from_:
                    # cycle
                    continue
                path.append(e)
                if edges_and_vertices:
                    path.append(v)
                if v.label == self.labelEnd:
                    yield path
                else:
                    for p in self.enumerate_all_paths(edges_and_vertices, path):
                        yield p

    def edit_distance_path(self, p1, p2, g1, g2,
                           function_mach_vertices=None,
                           function_match_edges=None,
                           use_min=False,
                           debug=False,
                           cache=None):
        """
        Tries to align two paths from two graphs.

        @param      p1                      path 1 (from g1)
        @param      p2                      path 2 (from g2)
        @param      g1                      graph 1
        @param      g2                      graph 2
        @param      function_mach_vertices  function which gives a distance bewteen two vertices,
                                            if None, it take the output of @see me get_matching_functions
        @param      function_match_edges    function which gives a distance bewteen two edges,
                                            if None, it take the output of @see me get_matching_functions
        @param      use_min                 the returned is based on a edit distance, if this parameter is True, the returned value will be:

                                            ::

                                                if use_min :
                                                    n = min (len(p1), len(p2))
                                                    d = d*1.0 / n if n > 0 else 0

        @param      debug                   unused
        @param      cache                   to cache the costs
        @return                             2-uple: distance, aligned path
        """
        def edge_vertex_match(x, y, g1, g2, w1, w2):
            if x is None:
                if y is None:
                    raise RuntimeError(
                        "Both x and y are None.")
                return y.weight * w2
            elif y is None:
                return x.weight * w1
            return (x.weight * w1 + y.weight * w2) / 2

        def cache_cost(func, a, b, g1, g2, w1, w2):
            if cache is None:
                return func(a, b, g1, g2, w1, w2)
            key = (id(a), id(b), w1, w2)
            if key in cache:
                return cache[key]
            cost = func(a, b, g1, g2, w1, w2)
            cache[key] = cost
            return cost

        function_mach_vertices, function_match_edges = (
            self.get_matching_functions(
                function_mach_vertices, function_match_edges, True))
        dist = {(-1, -1): (0, None, None)}

        if use_min:
            w1 = 1.0 / len(p1)
            w2 = 1.0 / len(p2)
        else:
            w1 = 1.
            w2 = 1.

        p2l = list(enumerate(p2))
        for i1, eorv1 in enumerate(p1):
            ed1 = eorv1.is_edge()
            ve1 = eorv1.is_vertex()
            for i2, eorv2 in p2l:
                np = i1, i2
                posit = [((i1 - 1, i2 - 1), (eorv1, eorv2)),
                         ((i1 - 1, i2), (eorv1, None)),
                         ((i1, i2 - 1), (None, eorv2))]

                if ed1 and eorv2.is_edge():
                    func = function_match_edges
                elif ve1 and eorv2.is_vertex():
                    func = function_mach_vertices
                else:
                    func = edge_vertex_match

                for p, co in posit:
                    if p in dist:
                        c0 = dist[p][0]
                        c1 = cache_cost(func, co[0], co[1], g1, g2, w1, w2)
                        c = c0 + c1
                        if np not in dist or c < dist[np][0]:
                            dist[np] = (c, p, co, (c0, c1))

        last = dist[len(p1) - 1, len(p2) - 1]
        path = []
        while last[1] is not None:
            path.append(last)
            last = dist[last[1]]

        path.reverse()

        d = dist[len(p1) - 1, len(p2) - 1][0]
        if use_min:
            n = min(len(p1), len(p2))
            d = d * 1.0 / n if n > 0 else 0
        return d, path

    def private_count_left_right(self, valuesInList):
        countLeft = {}
        countRight = {}
        for k, v in valuesInList:
            i, j = v
            if i not in countRight:
                countRight[i] = {}
            countRight[i][j] = countRight[i].get(j, 0) + 1
            if j not in countLeft:
                countLeft[j] = {}
            countLeft[j][i] = countLeft[j].get(i, 0) + 1
        return countLeft, countRight

    def private_kruskal_matrix(self, matrix, reverse):
        countLeft, countRight = self.private_count_left_right(matrix)
        cleft, cright = len(countLeft), len(countRight)
        matrix.sort(reverse=reverse)
        count = max(max([sum(_.values()) for _ in countRight.values()]),
                    max([sum(_.values()) for _ in countLeft.values()]))
        while count > 1:
            k, v = matrix.pop()
            i, j = v
            countRight[i][j] -= 1
            countLeft[j][i] -= 1
            count = max(max([max(_.values()) for _ in countRight.values()]),
                        max([max(_.values()) for _ in countLeft.values()]))

        mini = min(cleft, cright)
        if len(matrix) < mini:
            raise Exception("impossible: the smallest set should get all" +
                            "its element associated to at least one coming from the other set")

    def _private_string_path_matching(self, path, skipEdge=False):
        temp = []
        for p in path:
            u, v = p[2]
            if skipEdge and ((u is not None and u.is_edge()) or
                             (v is not None and v.is_edge())):
                continue
            su = "-" if u is None else str(u.nb)
            sv = "-" if v is None else str(v.nb)
            s = "(%s,%s)" % (su, sv)
            temp.append(s)
        return " ".join(temp)

    def distance_matching_graphs_paths(self, graph2,
                                       function_mach_vertices=None,
                                       function_match_edges=None,
                                       noClean=False,
                                       store=None,
                                       use_min=True,
                                       weight_vertex=1.,
                                       weight_edge=1.,
                                       verbose=0, fLOG=print):
        """
        Computes an alignment between two graphs.

        @param      graph2                  the other graph
        @param      function_mach_vertices  function which gives a distance bewteen two vertices,
                                            if None, it take the output of @see me get_matching_functions
        @param      function_match_edges    function which gives a distance bewteen two edges,
                                            if None, it take the output of @see me get_matching_functions
        @param      noClean                 if True, clean unmatched vertices and edges
        @param      store                   if None, does nothing, if it is a dictionary, the function will store here various
                                            information about how th matching was operated
        @param      use_min                 @see me edit_distance_path
        @param      weight_vertex           a weight for every vertex
        @param      weight_edge             a weight for every edge
        @param      verbose                 display some progress with :epkg:`tqdm`
        @param      fLOG                    logging functino
        @return                             2 tuple (a distance, a graph containing the aligned paths between the two graphs)

        See :ref:`l-graph_distance`.
        """

        function_mach_vertices, function_match_edges = (
            self.get_matching_functions(
                function_mach_vertices, function_match_edges, True))

        paths1 = list(self.enumerate_all_paths(True))
        paths2 = list(graph2.enumerate_all_paths(True))

        if store is not None:
            store["nbpath1"] = len(paths1)
            store["nbpath2"] = len(paths2)

        matrix_distance = {}
        if verbose > 0 and fLOG is not None:
            fLOG("[distance_matching_graphs_paths] builds matrix_distance")
            from tqdm import tqdm
            loop1 = tqdm(list(enumerate(paths1)))
        else:
            loop1 = enumerate(paths1)
        loop2 = list(enumerate(paths2))
        if verbose > 0 and fLOG is not None:
            fLOG("[distance_matching_graphs_paths] len(loop1)=%d" % len(
                list(enumerate(paths1))))
            fLOG("[distance_matching_graphs_paths] len(loop2)=%d" % len(loop2))
        cache = {}
        for i1, p1 in loop1:
            for i2, p2 in loop2:
                matrix_distance[i1, i2] = self.edit_distance_path(
                    p1, p2, self, graph2, function_mach_vertices,
                    function_match_edges, use_min=use_min,
                    cache=cache)
        if verbose > 0 and fLOG is not None:
            fLOG("[distance_matching_graphs_paths] len(cache)=%d" % len(cache))

        if store is not None:
            store["matrix_distance"] = matrix_distance
        reduction = [(v[0], k) for k, v in matrix_distance.items()]
        if store is not None:
            store["path_mat1"] = copy.deepcopy(reduction)
        if verbose > 0 and fLOG is not None:
            fLOG("[distance_matching_graphs_paths] private_kruskal_matrix")
        self.private_kruskal_matrix(reduction, False)
        if store is not None:
            store["path_mat2"] = copy.deepcopy(reduction)

        if verbose > 0 and fLOG is not None:
            fLOG("[distance_matching_graphs_paths] pair_count_vertex")
        pair_count_edge = {}
        pair_count_vertex = {}
        for k, v in reduction:
            path = matrix_distance[v][1]
            for el in path:
                n1, n2 = el[2]
                if n1 is not None and n2 is not None:
                    if n1.is_edge() and n2.is_edge():
                        add = n1.nb, n2.nb
                        pair_count_edge[add] = pair_count_edge.get(add, 0) + 1
                    elif n1.is_vertex() and n2.is_vertex():
                        add = n1.nb, n2.nb
                        pair_count_vertex[
                            add] = pair_count_vertex.get(add, 0) + 1

        if store is not None:
            store["pair_count_vertex"] = pair_count_vertex
            store["pair_count_edge"] = pair_count_edge

        reduction_edge = [(v, k) for k, v in pair_count_edge.items()]
        if store is not None:
            store["edge_mat1"] = copy.copy(reduction_edge)
        self.private_kruskal_matrix(reduction_edge, True)
        if store is not None:
            store["edge_mat2"] = copy.copy(reduction_edge)

        reduction_vertex = [(v, k) for k, v in pair_count_vertex.items()]
        if store is not None:
            store["vertex_mat1"] = copy.copy(reduction_vertex)
        self.private_kruskal_matrix(reduction_vertex, True)
        if store is not None:
            store["vertex_mat2"] = copy.copy(reduction_vertex)

        if verbose > 0 and fLOG is not None:
            fLOG("[distance_matching_graphs_paths] private_count_left_right")
        count_edge_left, count_edge_right = self.private_count_left_right(
            reduction_edge)
        count_vertex_left, count_vertex_right = self.private_count_left_right(
            reduction_vertex)

        res_graph = GraphDistance([])
        doneVertex = {}
        done_edge = {}

        if verbose > 0 and fLOG is not None:
            fLOG("[distance_matching_graphs_paths] builds merged graph")
        for k, v in self.vertices.items():
            newv = Vertex(v.nb, v.label, weight_vertex)
            res_graph.vertices[k] = newv
            if v.nb in count_vertex_right:
                ind = list(count_vertex_right[v.nb].keys())[0]
                newv.pair = (v, graph2.vertices[ind])
                doneVertex[ind] = newv
                if newv.pair[0].label != newv.pair[1].label:
                    newv.label = "%s|%s" % (
                        newv.pair[0].label, newv.pair[1].label)
            else:
                newv.pair = (v, None)

        for k, v in graph2.vertices.items():
            if k in doneVertex:
                continue
            newv = Vertex("2a.%s" % v.nb, v.label, weight_vertex)
            res_graph.vertices[newv.nb] = newv
            newv.pair = (None, v)

        for k, e in self.edges.items():
            newe = Edge(e.from_, e.to, e.label, weight_edge)
            res_graph.edges[k] = newe
            if e.nb in count_edge_right:
                ind = list(count_edge_right[e.nb].keys())[0]
                newe.pair = (e, graph2.edges[ind])
                done_edge[ind] = newe
            else:
                newe.pair = (e, None)

        for k, e in graph2.edges.items():
            if k in done_edge:
                continue
            from_ = list(count_vertex_left[e.from_].keys())[0] if e.from_ in count_vertex_left \
                else "2a.%s" % e.from_
            to = list(count_vertex_left[e.to].keys())[0] if e.to in count_vertex_left \
                else "2a.%s" % e.to
            if from_ not in res_graph.vertices:
                raise RuntimeError("should not happen " +
                                   from_)  # pragma: no cover
            if to not in res_graph.vertices:
                raise RuntimeError("should not happen " +
                                   to)  # pragma: no cover
            newe = Edge(from_, to, e.label, weight_edge)
            res_graph.edges[newe.nb] = newe  # pylint: disable=E1101
            newe.pair = (None, e)

        if verbose > 0 and fLOG is not None:
            fLOG(
                "[distance_matching_graphs_paths] compute_predecessor, compute_successor")
        res_graph.compute_predecessor()
        res_graph.compute_successor()

        allPaths = list(res_graph.enumerate_all_paths(True))

        temp = [sum([0 if None in _.pair else 1 for _ in p]) * 1.0 / len(p)
                for p in allPaths]
        distance = 1.0 - 1.0 * sum(temp) / len(allPaths)

        return distance, res_graph

    def draw_vertices_edges(self):
        vertices = []
        edges = []
        for k, v in self.vertices.items():
            if v.pair == (None, None) or (v.pair[0] is not None and v.pair[1] is not None):
                vertices.append((k, v.label))
            elif v.pair[1] is None:
                vertices.append((k, "-" + v.label, "red"))
            elif v.pair[0] is None:
                vertices.append((k, "+" + v.label, "green"))
            else:
                raise RuntimeError("?")  # pragma: no cover

        for k, v in self.edges.items():
            if v.pair == (None, None) or (v.pair[0] is not None and v.pair[1] is not None):
                edges.append((v.from_, v.to, v.label))
            elif v.pair[1] is None:
                edges.append((v.from_, v.to, "-" + v.label, "red"))
            elif v.pair[0] is None:
                edges.append((v.from_, v.to, "+" + v.label, "green"))
            else:
                raise RuntimeError("?")  # pragma: no cover

        return vertices, edges



"""
@file
@brief About completion, simple algorithm
"""
import time
from typing import Tuple, List, Iterator, Dict
from pyquickhelper.loghelper import noLOG
from .completion import CompletionTrieNode


class CompletionElement:
    """
    Definition of an element in a completion system,
    it contains the following members:

    * *value*: the completion
    * *weight*: a weight or a position, we assume a completion with
      a lower weight is shown at a lower position
    * *disp*: display string (no impact on the algorithm)

    * mks0*: value of minimum keystroke
    * mks0_*: length of the prefix to obtain *mks0*

    * *mks1*: value of dynamic minimum keystroke
    * *mks1_*: length of the prefix to obtain *mks1*

    * *mks2*: value of modified dynamic minimum keystroke
    * *mks2_*: length of the prefix to obtain *mks2*
    """

    __slots__ = "value", "weight", "disp", 'mks0', 'mks0_', 'mks1', 'mks1_', 'mks2', 'mks2_', 'prefix', '_info'

    def __init__(self, value: str, weight=1.0, disp=None):
        """
        constructor

        @param      value       value (a character)
        @param      weight      ordering (the lower, the first)
        @param      disp        original string, use this to identify the node
        """
        if not isinstance(value, str):
            raise TypeError(
                "value must be str not '{0}' - type={1}".format(value, type(value)))
        self.value = value
        self.weight = weight
        self.disp = disp
        self.prefix = None
        self._info = None

    @staticmethod
    def empty_prefix():
        """
        return an instance filled with an empty prefix
        """
        if not hasattr(CompletionElement, "static_empty_prefix"):
            res = CompletionElement('', None)
            res.mks0 = res.mks1 = res.mks2 = 0
            res.mks0_ = res.mks1_ = res.mks2_ = 0
            CompletionElement.static_empty_prefix = res
            return res
        else:
            return CompletionElement.static_empty_prefix

    def __repr__(self):
        """
        usual
        """
        if self._info:
            return "CompletionElementInfo('{0}'{1}{2})".format(self.value,
                                                               ", {0}".format(
                                                                   self.weight) if self.weight != 1 else "",
                                                               ", disp='{0}'" if self.disp else "")
        else:
            return "CompletionElement('{0}'{1}{2})".format(self.value,
                                                           ", {0}".format(
                                                               self.weight) if self.weight != 1 else "",
                                                           ", disp='{0}'" if self.disp else "")

    def str_mks0(self) -> str:
        """
        return a string with metric information
        """
        if hasattr(self, "mks0"):
            return "MKS={0} *={1}".format(self.mks0, self.mks0_)
        else:
            return "-"

    def str_mks(self) -> str:
        """
        return a string with metric information
        """
        s0 = self.str_mks0()
        if hasattr(self, "mks1"):
            return s0 + " |'={0} *={1} |\"={2} *={3}".format(
                self.mks1, self.mks1_, self.mks2, self.mks2_)
        else:
            return s0

    def str_all_completions(self, maxn=10, use_precompute=True) -> str:
        """
        builds a string with all completions for all
        prefixes along the paths, this is only available
        if parameter *completions* was used when calling
        method @see me update_metrics.

        @param      maxn            maximum number of completions to show
        @param      use_precompute  use intermediate results built by @see me precompute_stat
        @return                     str
        """
        rows = [
            "{0} -- {1} -- {2}".format(self.weight, self.value, self.str_mks())]
        if self._info is not None:
            rows.append("------------------")
            for el in self._info._log_imp:
                rows.append(str(el))
            for i in range(len(self.value)):
                prefix = self.value[:i]
                rows.append("------------------")
                rows.append("i={0} - {1}".format(i, prefix))
                completions = self._info._completions.get(prefix, [])
                for i2, el in enumerate(completions):
                    ar = "   " if el.value != self.value else "-> "
                    add = "{5}{0}:{1} -- {2}{4}-- {3}".format(
                          i2, el.weight, el.value, el.str_mks(), " " * (20 - len(el.value)), ar)
                    rows.append(add)
        else:
            rows.append("NO INFO")
        return "\n".join(rows)

    def init_metrics(self, position: int, completions: List['CompletionElement'] = None):
        """
        initiate the metrics

        @param      position    position in the completion system when prefix is null,
                                *position starting from 0*
        @param      completions displayed completions, if not None, the method will
                                store them in member *_completions*
        @return                 boolean which indicates there was an update
        """
        if completions is not None:
            log_imp = True

            class c:
                def __str__(self):
                    return "{0}-{1}".format(self._completions, self._log_imp)

            self._info = c()
            self._info._completions = {}
            self._info._log_imp = []
            if '' not in self._info._completions:
                cut = min(15, max(10, len(self.value)),
                          len(completions['']))
                if len(completions['']) >= cut:
                    self._info._completions[''] = completions[''][:cut]
                else:
                    self._info._completions[''] = completions[''].copy()
        else:
            log_imp = False

        self.prefix = CompletionElement.empty_prefix()
        position += 1
        if len(self.value) <= position:
            self.mks0 = len(self.value)
            self.mks1 = len(self.value)
            self.mks2 = len(self.value)
            self.mks0_ = len(self.value)
            self.mks1_ = len(self.value)
            self.mks2_ = len(self.value)
            return False
        else:
            self.mks0 = position
            self.mks1 = position
            self.mks2 = position
            self.mks0_ = 0
            self.mks1_ = 0
            self.mks2_ = 0
            if log_imp:
                self._info._log_imp.append(
                    (0, "mks0", position, '', "k={0}".format(0),
                     "p={0}".format(position), "it={0}".format(0)))
            return True

    def update_metrics(self, prefix: str, position: int, improved: dict, delta: float,
                       completions: List['CompletionElement'] = None, iteration=-1):
        """
        update the metrics

        @param      prefix      prefix
        @param      position    position in the completion system when prefix has length k,
                                *position starting from 0*
        @param      improved    if one metrics is < to the completion length, it means
                                it can be used to improve others queries
        @param      delta       delta in the dynamic modified mks
        @param      completions displayed completions, if not None, the method will
                                store them in member *_completions*
        @param      iteration   for debugging purpose, indicates when this improvment was detected
        @return                 boolean which indicates there was an update
        """
        if self.prefix is not None and len(prefix) < len(self.prefix.value):
            # no need to look into it
            return False

        if completions is not None:
            log_imp = True
            if prefix not in self._info._completions:
                cut = min(15, max(10, len(self.value)),
                          len(completions[prefix]))
                if len(completions[prefix]) >= cut:
                    self._info._completions[prefix] = completions[prefix][:cut]
                else:
                    self._info._completions[
                        prefix] = completions[prefix].copy()
        else:
            log_imp = False

        k = len(prefix)
        pos = position + 1
        mks = k + pos
        check = False
        if mks < self.mks0:
            self.mks0 = mks
            self.mks0_ = k
            check = True
            if log_imp:
                self._info._log_imp.append(
                    (1, "mks0", mks, prefix, "k={0}".format(k), "p={0}".format(position),
                     "it={0}".format(iteration), "last={0}".format(self.prefix.value)))
        elif mks == self.mks0 and self.mks0_ < k:
            self.mks0_ = k
        if mks < self.mks1:
            self.mks1 = mks
            self.mks1_ = k
            check = True
        if mks < self.mks2:
            self.mks2 = mks
            self.mks2_ = k
            check = True

        if self.prefix and len(self.prefix.value) < len(prefix):
            # we use the latest prefix available
            v = self.prefix
            dd = len(prefix) - len(v.value) + pos
            mks = v.mks1 + dd
            if mks < self.mks1:
                self.mks1 = mks
                self.mks1_ = k
                check = True
                if log_imp:
                    self._info._log_imp.append(
                        (4, "mks1", mks, prefix, "k={0}".format(k), "p={0}".format(position),
                         "it={0}".format(iteration), "last={0}".format(self.prefix.value)))
            mks = v.mks2 + dd
            if mks < self.mks2:
                self.mks2 = mks
                self.mks2_ = k
                check = True
                if log_imp:
                    self._info._log_imp.append(
                        (5, "mks2", mks, prefix, "k={0}".format(k), "p={0}".format(position),
                         "it={0}".format(iteration), "last={0}".format(self.prefix.value)))
        if prefix in improved:
            v = improved[prefix]
            self.prefix = v
            mks = v.mks1 + min(len(self.value) - len(prefix), pos)
            if mks < self.mks1:
                self.mks1 = mks
                self.mks1_ = k
                check = True
                if log_imp:
                    self._info._log_imp.append(
                        (2, "mks1", mks, prefix, "k={0}".format(k), "p={0}".format(position),
                         "it={0}".format(iteration), "last={0}".format(self.prefix.value)))
            mks = v.mks2 + min(len(self.value) - len(prefix), pos + delta)
            if mks < self.mks2:
                self.mks2 = mks
                self.mks2_ = k
                check = True
                if log_imp:
                    self._info._log_imp.append(
                        (3, "mks2", mks, prefix, "k={0}".format(k), "p={0}".format(position),
                         "it={0}".format(iteration), "last={0}".format(self.prefix.value)))

        if log_imp and self.prefix and self.prefix.value != '':
            self._info._log_imp.append(self.prefix)
        return check


class CompletionSystem:
    """
    define a completion system
    """

    def __init__(self, elements: List[CompletionElement]):
        """
        fill the completion system
        """
        def create_element(i, e):
            if isinstance(e, CompletionElement):
                return e
            if isinstance(e, tuple):
                return CompletionElement(e[1], e[0] if e[0] else i)
            return CompletionElement(e, i)
        self._elements = [create_element(i, e) for i, e in enumerate(elements)]

    def __getitem__(self, i):
        """
        Returns ``elements[i]``.
        """
        return self._elements[i]

    def find(self, value: str, is_sorted=False) -> CompletionElement:
        """
        Not very efficient, finds an item in a the list.

        @param      value       string to find
        @param      is_sorted   the function will assume the elements are sorted by
                                alphabetical order
        @return                 element or None
        """
        if is_sorted:
            raise NotImplementedError(  # pragma: no cover
                "No optimisation for the sorted case.")
        for e in self:
            if e.value == value:
                return e
        return None

    def items(self) -> Iterator[Tuple[str, CompletionElement]]:
        """
        Iterates on ``(e.value, e)``.
        """
        for e in self._elements:
            yield e.value, e

    def tuples(self) -> Iterator[Tuple[float, str]]:
        """
        Iterates on ``(e.weight, e.value)``.
        """
        for e in self._elements:
            yield e.weight, e.value

    def __len__(self) -> int:
        """
        Number of elements.
        """
        return len(self._elements)

    def __iter__(self) -> Iterator[CompletionElement]:
        """
        Iterates over elements.
        """
        for e in self._elements:
            yield e

    def sort_values(self):
        """
        sort the elements by value
        """
        self._elements = list(
            _[-1] for _ in sorted((e.value, e.weight, e) for e in self))

    def sort_weight(self):
        """
        Sorts the elements by value.
        """
        self._elements = list(
            _[-1] for _ in sorted((e.weight, e.value, e) for e in self))

    def compare_with_trie(self, delta=0.8, fLOG=noLOG):
        """
        Compares the results with the other implementation.

        @param      delta       parameter *delta* in the dynamic modified mks
        @param      fLOG        logging function
        @return                 None or differences
        """
        def format_diff(el, f, diff):
            s = "VALUE={0}\nSYST=[{1}]\nTRIE=[{2}]\nMORE SYSTEM:\n{3}\n######\nMORE TRIE:\n{4}".format(
                el.value, el.str_mks(), f.stat.str_mks(),
                el.str_all_completions(), f.str_all_completions())
            if diff:
                return "-------\n{0}\n-------".format(s)
            return s

        trie = CompletionTrieNode.build(self.tuples())
        self.compute_metrics(delta=delta, fLOG=fLOG, details=True)
        trie.precompute_stat()
        trie.update_stat_dynamic(delta=delta)
        diffs = []
        for el in self:
            f = trie.find(el.value)
            d0 = el.mks0 - f.stat.mks0
            d1 = el.mks1 - f.stat.mks1
            d2 = el.mks2 - f.stat.mks2
            d4 = el.mks0_ - f.stat.mks0_
            if d0 != 0 or d1 != 0 or d2 != 0 or d4 != 0:
                diffs.append((d0, d1, d2, d4, el, f, format_diff(el, f, True)))
        if diffs:
            diffs.sort(key=str)
            return diffs
        else:
            return None

    def to_dict(self) -> Dict[str, CompletionElement]:
        """
        Returns a dictionary.
        """
        return {el.value: el for el in self}

    def compute_metrics(self, ffilter=None, delta=0.8,
                        details=False, fLOG=noLOG) -> int:
        """
        Computes the metric for the completion itself.

        @param      ffilter     filter function
        @param      delta       parameter *delta* in the dynamic modified mks
        @param      details     log more details about displayed completions
        @param      fLOG        logging function
        @return                 number of iterations

        The function ends by sorting the set of completion by alphabetical order.
        """
        self.sort_weight()
        if ffilter is not None:
            raise NotImplementedError(  # pragma: no cover
                "ffilter not None is not implemented")
        if details:
            store_completions = {'': []}

        improved = {}
        to = time.perf_counter()
        fLOG("init_metrics:", len(self))
        for i, el in enumerate(self._elements):
            if details:
                store_completions[''].append(el)
                r = el.init_metrics(i, store_completions)
            else:
                r = el.init_metrics(i)
            if r and el.value not in improved:
                improved[el.value] = el
        t = time.perf_counter()
        fLOG(
            "interation 0: #={0} dt={1} - log details={2}".format(len(self), t - to, details))

        updates = 1
        it = 1
        while updates > 0:
            displayed = {}
            updates = 0
            for i, el in enumerate(self._elements):
                for k in range(0, len(el.value)):
                    prefix = el.value[:k]
                    if prefix not in displayed:
                        displayed[prefix] = 0
                        if details:
                            store_completions[prefix] = [el]
                    else:
                        displayed[prefix] += 1
                        if details:
                            store_completions[prefix].append(el)
                    r = el.update_metrics(
                        prefix, displayed[prefix], improved, delta,
                        completions=(store_completions if details else None),
                        iteration=it)
                    if r:
                        if el.value not in improved:
                            improved[el.value] = el
                        updates += 1
            t = time.perf_counter()
            fLOG("interation {0}: updates={1} dt={2}".format(
                it, updates, t - to))
            it += 1

        self.sort_values()
        return it - 1

    def enumerate_test_metric(self, qset: Iterator[Tuple[str, float]]) -> Iterator[Tuple[CompletionElement, CompletionElement]]:
        """
        Evaluates the completion set on a set of queries,
        the function returns a list of @see cl CompletionElement
        with the three metrics :math:`M`, :math:`M'`, :math:`M"`
        for these particular queries.

        @param      qset        list of tuple(str, float) = (query, weight)
        @return                 list of tuple of @see cl CompletionElement,
                                the first one is the query, the second one is the None or
                                the matching completion

        The method @see me compute_metric needs to be called first.
        """
        qset = sorted(qset)
        current = 0
        for query, weight in qset:
            while current < len(self) and self[current].value <= query:
                current += 1
            ind = current - 1
            el = CompletionElement(query, weight)
            if ind >= 0:
                inset = self[ind]
                le = len(inset.value)
                if le <= len(query) and inset.value == query[:le]:
                    if le == len(query):
                        found = inset
                        el.mks0 = inset.mks0
                        el.mks1 = inset.mks1
                        el.mks2 = inset.mks2
                        el.mks0_ = len(query)
                        el.mks1_ = len(query)
                        el.mks2_ = len(query)
                    else:
                        found = None
                        el.mks0 = 0
                        el.mks0_ = 0
                        el.mks1 = inset.mks1 + len(query) - le
                        el.mks1_ = le
                        el.mks2 = inset.mks2 + len(query) - le
                        el.mks2_ = le
                else:
                    found = None
                    el.mks0 = len(query)
                    el.mks1 = len(query)
                    el.mks2 = len(query)
                    el.mks0_ = len(query)
                    el.mks1_ = len(query)
                    el.mks2_ = len(query)
            else:
                found = None
                el.mks0 = len(query)
                el.mks1 = len(query)
                el.mks2 = len(query)
                el.mks0_ = len(query)
                el.mks1_ = len(query)
                el.mks2_ = len(query)

            yield el, found

    def test_metric(self, qset: Iterator[Tuple[str, float]]) -> Dict[str, float]:
        """
        Evaluates the completion set on a set of queries,
        the function returns a dictionary with the aggregated metrics and
        some statistics about them.

        @param      qset        list of tuple(str, float) = (query, weight)
        @return                 list of @see cl CompletionElement

        The method @see me compute_metric needs to be called first.
        It then calls @see me enumerate_metric.
        """
        res = dict(mks0=0.0, mks1=0.0, mks2=0.0,
                   sum_weights=0.0, sum_wlen=0.0, n=0)
        hist = {k: {} for k in {"mks0", "mks1", "mks2", "l"}}
        wei = {k: {} for k in hist}
        res["hist"] = hist
        res["histnow"] = wei

        for el, _ in self.enumerate_test_metric(qset):
            le = len(el.value)
            w = el.weight
            res["mks0"] += w * el.mks0
            res["mks1"] += w * el.mks1
            res["mks2"] += w * el.mks2
            res["sum_weights"] += w
            res["sum_wlen"] += w * le
            res["n"] += 1

            if el.mks0 not in hist["mks0"]:
                hist["mks0"][el.mks0] = w
                wei["mks0"][el.mks0] = 1
            else:
                hist["mks0"][el.mks0] += w
                wei["mks0"][el.mks0] += 1
            if el.mks1 not in hist["mks1"]:
                hist["mks1"][el.mks1] = w
                wei["mks1"][el.mks1] = 1
            else:
                hist["mks1"][el.mks1] += w
                wei["mks1"][el.mks1] += 1
            if el.mks2 not in hist["mks2"]:
                hist["mks2"][el.mks2] = w
                wei["mks2"][el.mks2] = 1
            else:
                hist["mks2"][el.mks2] += w
                wei["mks2"][el.mks2] += 1
            if le not in hist["l"]:
                hist["l"][le] = w
                wei["l"][le] = 1
            else:
                hist["l"][le] += w
                wei["l"][le] += 1
        return res



"""
@file
@brief graphviz helper
"""
import os
import sys
from pyquickhelper.loghelper import run_cmd
from pyquickhelper.helpgen.conf_path_tools import find_graphviz_dot


def run_graphviz(filename, image, engine="dot"):
    """
    Run :epkg:`GraphViz`.

    @param      filename        filename which contains the graph definition
    @param      image           output image
    @param      engine          *dot* or *neato*
    @return                     output of graphviz
    """
    ext = os.path.splitext(image)[-1]
    if ext != ".png":
        raise Exception("extension should be .png not " + str(ext))
    if sys.platform.startswith("win"):
        bin_ = os.path.dirname(find_graphviz_dot())
        # if bin not in os.environ["PATH"]:
        #    os.environ["PATH"] = os.environ["PATH"] + ";" + bin
        cmd = '"{0}\\{3}" -Tpng "{1}" -o "{2}"'.format(
            bin_, filename, image, engine)
    else:
        cmd = '"{0}" -Tpng "{1}" -o "{2}"'.format(engine, filename, image)
    out, err = run_cmd(cmd, wait=True)
    if len(err) > 0:
        raise Exception(
            "Unable to run Graphviz\nCMD:\n{0}\nOUT:\n{1}\nERR:\n{2}".format(cmd, out, err))
    return out


def edges2gv(vertices, edges):
    """
    Converts a graph into a :epkg:`GraphViz` file format.

    @param      edges           see below
    @param      vertices        see below
    @return                     gv format

    The function creates a file ``<image>.gv``.

    .. runpython::
        :showcode:

        from mlstatpy.graph.graphviz_helper import edges2gv
        gv = edges2gv([(1, "eee", "red")],
                      [(1, 2, "blue"), (3, 4), (1, 3)])
        print(gv)

    """
    memovertex = {}
    for v in vertices:
        if isinstance(v, tuple):
            if len(v) == 1:
                memovertex[v[0]] = None
            else:
                memovertex[v[0]] = v[1:]
        else:
            memovertex[v] = None
    for edge in edges:
        i, j = edge[:2]
        if i not in memovertex:
            memovertex[i] = None
        if j not in memovertex:
            memovertex[j] = None

    li = ["digraph{"]
    for k, v in memovertex.items():
        if v is None:
            li.append("%s ;" % k)
        elif len(v) == 1:
            li.append("\"%s\" [label=\"%s\"];" % (k, v[0]))
        elif len(v) == 2:
            li.append("\"%s\" [label=\"%s\",fillcolor=%s,color=%s];" % (
                k, v[0], v[1], v[1]))
        else:
            raise ValueError("unable to understand " + str(v))

    for edge in edges:
        i, j = edge[:2]
        if len(edge) == 2:
            li.append("\"%s\" -> \"%s\";" % (i, j))
        elif len(edge) == 3:
            li.append("\"%s\" -> \"%s\" [label=\"%s\"];" % (i, j, edge[2]))
        elif len(edge) == 4:
            li.append(
                "\"%s\" -> \"%s\" [label=\"%s\",color=%s];" % (i, j, edge[2], edge[3]))
        else:
            raise ValueError("unable to understand " + str(edge))
    li.append("}")

    text = "\n".join(li)
    return text


def draw_graph_graphviz(vertices, edges, image=None, engine="dot"):
    """
    Draws a graph using :epkg:`Graphviz`.

    @param      edges           see below
    @param      vertices        see below
    @param      image           output image, None, just returns the output
    @param      engine          *dot* or *neato*
    @return                     :epkg:`Graphviz` output or
                                the dot text if *image* is None

    The function creates a file ``<image>.gv`` if *image* is not None.
    ::

        edges    = [ (1,2, label, color), (3,4), (1,3), ... ]  , liste d'arcs
        vertices = [ (1, label, color), (2), ... ]  , liste de noeuds
        image = nom d'image (format png)

    """
    text = edges2gv(vertices, edges)
    if image is None:
        return text
    filename = image + ".gv"
    with open(filename, "w", encoding="utf-8") as f:
        f.write(text)

    out = run_graphviz(filename, image, engine=engine)
    if not os.path.exists(image):
        raise FileNotFoundError(
            "GraphViz failed with no reason. '{0}' not found.".format(image))
    return out



# -*- coding: utf-8 -*-
"""
@file
@brief Ce module définit un segment qui va parcourir l'image,
en plus d'être un segment, cette classe inclut la dimension de l'image,
et une fonction repérant sur ce segment les gradients presque
orthogonaux à l'image.
"""
import copy
import numpy
from .geometrie import Segment, Point


class SegmentBord_Commun(Segment):
    """
    Définit un segment allant d'un bord a un autre de l'image,
    la méthode importante est @see me decoupe_gradient.

        dim     est la dimension de l'image"""

    # voir la remarque dans la classe Point a propos de __slots__
    __slots__ = ("dim",)

    def __init__(self, dim):
        """constructeur, definit la definition de l'image"""
        Segment.__init__(self, Point(0, 0), Point(0, 0))
        self.dim = dim

    def copy(self):
        """
        Copie l'instance.
        """
        return copy.deepcopy(self)

    def __str__(self):
        """permet d'afficher le segment"""
        s = Segment.__str__(self)
        s += " -- dim -- " + self.dim.__str__()
        return s

    def decoupe_gradient(self, gradient, cos_angle, ligne_gradient, seuil_norme):
        """
        Pour un segment donne joignant deux bords de l'image,
        cette fonction récupère le gradient et construit une liste
        contenant des informations pour un pixel sur deux du segment,

        * norme* : mémorise la norme du gradient en ce point de l'image
        * *pos* : mémorise la position du pixel
        * *aligne* : est vrai si le gradient est presque orthogonale au segment,
          ce resultat est relié au paramètre proba_bin,
          deux vecteurs sont proches en terme de direction,
          s'ils font partie du secteur angulaire défini par *proba_bin*.

        Le parcours du segment commence à son origine ``self.a``,
        et on ajoute à chaque itération deux fois le vecteur normal
        jusqu'à sortir du cadre de l'image,
        les informations sont stockées dans ``ligne_gradient`` qui a une liste
        d'informations préalablement créée au debut du programme
        de facon à gagner du temps.
        """
        n = self.directeur()
        nor = self.normal().as_array()
        n.scalairek(2.0)
        p = copy.copy(self.a)
        a = p.arrondi()

        i = 0
        while a.x >= 0 and a.y >= 0 and a.x < self.dim.x and a.y < self.dim.y:
            # on recupere l'élément dans ligne ou doivent être
            # stockées les informations (ligne_gradient)
            t = ligne_gradient.info_ligne[i]

            # on recupere le gradient de l'image au pixel a
            g = gradient[a.y, a.x]

            # on calcul sa norme
            t.norme = (g[0] ** 2 + g[1]**2) ** 0.5

            # on place les coordonnees du pixel dans t
            t.pos.x = p.x
            t.pos.y = p.y

            # si la norme est positive, le gradient à une direction
            # on regarde s'il est dans le meme secteur angulaire (proba_bin)
            # que le vecteur normal au segment (nor)
            if t.norme > seuil_norme:
                t.aligne = numpy.dot(g, nor) > cos_angle * t.norme
            else:
                t.aligne = False

            # on passe au pixel suivant
            p += n
            a = p.arrondi()   # calcul de l'arrondi
            i += 1

        # on indique a ligne_gradient le nombre de pixel pris en compte
        # ensuite, on decidera si ce segment est effectivement un segment de l'image
        ligne_gradient.nb = i



# -*- coding: utf-8 -*-
"""
@file
@brief Conversion from tree to neural network.
"""
import numpy
from ..optim import SGDOptimizer


class _TrainingAPI:
    """
    Declaration of function needed to train a model.
    """

    @property
    def training_weights(self):
        "Returns the weights."
        raise NotImplementedError(  # pragma: no cover
            "This should be overwritten.")

    def update_training_weights(self, grad, add=True):
        """
        Updates weights.

        :param grad: vector to add to the weights such as gradient
        :param add: addition or replace
        """
        raise NotImplementedError(  # pragma: no cover
            "This should be overwritten.")

    def fill_cache(self, X):
        """
        Creates a cache with intermediate results.
        """
        return None  # pragma: no cover

    def loss(self, X, y, cache=None):
        """
        Computes the loss. Returns a float.
        """
        raise NotImplementedError(  # pragma: no cover
            "This should be overwritten.")

    def dlossds(self, X, y, cache=None):
        """
        Computes the loss derivative due to prediction error.
        """
        raise NotImplementedError(  # pragma: no cover
            "This should be overwritten.")

    def gradient_backward(self, graddx, X, inputs=False, cache=None):
        """
        Computes the gradient in X.

        :param graddx: existing gradient against the outputs
        :param X: computes the gradient in X
        :param inputs: if False, derivative against the coefficients,
            otherwise against the inputs.
        :param cache: cache intermediate results to avoid more computation
        :return: gradient
        """
        raise NotImplementedError(  # pragma: no cover
            "This should be overwritten.")

    def gradient(self, X, y, inputs=False):
        """
        Computes the gradient in *X* knowing the expected value *y*.

        :param X: computes the gradient in X
        :param y: expected values
        :param inputs: if False, derivative against the coefficients,
            otherwise against the inputs.
        :return: gradient
        """
        if len(X.shape) != 1:
            raise ValueError(  # pragma: no cover
                "X must a vector of one dimension but has shape {}.".format(X.shape))
        cache = self.fill_cache(X)  # pylint: disable=E1128
        dlossds = self.dlossds(X, y, cache=cache)
        return self.gradient_backward(dlossds, X, inputs=inputs, cache=cache)

    def fit(self, X, y, optimizer=None, max_iter=100, early_th=None, verbose=False,
            lr=None, lr_schedule=None, l1=0., l2=0., momentum=0.9):
        """
        Fits a neuron.

        :param X: training set
        :param y: training labels
        :param optimizer: optimizer, by default, it is
            :class:`SGDOptimizer <mlstatpy.optim.sgd.SGDOptimizer>`.
        :param max_iter: number maximum of iterations
        :param early_th: early stopping threshold
        :param verbose: more verbose
        :param lr: to overwrite *learning_rate_init* if
            *optimizer* is None (unused otherwise)
        :param lr_schedule: to overwrite *lr_schedule* if
            *optimizer* is None (unused otherwise)
        :param l1: L1 regularization if *optimizer* is None
            (unused otherwise)
        :param l2: L2 regularization if *optimizer* is None
            (unused otherwise)
        :param momentum: used if *optimizer* is None
        :return: self
        """
        if optimizer is None:
            optimizer = SGDOptimizer(
                self.training_weights, learning_rate_init=lr or 0.002,
                lr_schedule=lr_schedule or 'invscaling',
                l1=l1, l2=l2, momentum=momentum)

        def fct_loss(coef, lx, ly, neuron=self):
            neuron.update_training_weights(coef, False)
            loss = neuron.loss(lx, ly)
            if loss.shape[0] > 1:
                return numpy.sum(loss)
            return loss

        def fct_grad(coef, lx, ly, i, neuron=self):
            neuron.update_training_weights(coef, False)
            return neuron.gradient(lx, ly).ravel()

        optimizer.train(
            X, y, fct_loss, fct_grad, max_iter=max_iter,
            early_th=early_th, verbose=verbose)

        self.update_training_weights(optimizer.coef, False)
        return self



# -*- coding: utf-8 -*-
"""
@file
@brief Text normalization
"""
import unicodedata


def remove_diacritics(input_str):
    """
    remove diacritics

    @param      input_str       string to clean
    @return                     cleaned string

    Example::

        enguérand --> enguerand

    .. versionadded:: 1.0
    """
    nkfd_form = unicodedata.normalize('NFKD', input_str)
    only_ascii = nkfd_form.encode('ASCII', 'ignore')
    return only_ascii.decode("utf8")



# -*- coding: utf-8 -*-
"""
@file
@brief Définition de petits éléments géométriques tels que les points
et les segments, implemente également des opérations standard
telles le produit scalaire entre deux vecteurs, ...
"""
import math
import copy
import numpy


class Point:
    """
    Définit un point de l'image ou un vecteur,
    deux coordonnées *x* et *y* qui sont réelles.
    """
    __slots__ = "x", "y"

    def __init__(self, x, y):
        """constructeur"""
        self.x = x
        self.y = y

    def __str__(self):
        """permet d'afficher un point avec l'instruction print"""
        return '({0},{1})'.format(self.x, self.y)

    def __repr__(self):
        """usuel"""
        return 'Point({0}, {1})'.format(self.x, self.y)

    def normalise(self):
        """normalise le vecteur, sa norme devient 1"""
        v = self.x * self.x + self.y * self.y
        v = math.sqrt(v)
        if v > 0:   # evite les erreurs si sa norme est nulle
            self.x /= v
            self.y /= v

    def scalairek(self, k: float):
        """
        Mulitplication par un scalaire.

        @param      k       float
        """
        self.x *= k
        self.y *= k

    def norme(self) -> float:
        """
        Retourne la norme.

        @return         float (norm)
        """
        return math.sqrt(self.x * self.x + self.y * self.y)

    def as_array(self):
        """
        Convertit en array.
        """
        return numpy.array([self.x, self.y])

    def scalaire(self, k: 'Point') -> float:
        """
        Calcule le produit scalaire.

        @param      k       @see cl Point
        @return             float
        """
        return self.x * k.x + self.y * k.y

    def __iadd__(self, ad):
        """ajoute un vecteur à celui-ci"""
        self.x += ad.x
        self.y += ad.y
        return self

    def __add__(self, ad):
        """ajoute un vecteur a celui-ci"""
        return Point(self.x + ad.x, self.y + ad.y)

    def arrondi(self) -> 'Point':
        """
        retourne les coordonnées arrondies à l'entier le plus proche
        """
        return Point(int(self.x + 0.5), int(self.y + 0.5))

    def __sub__(self, p):
        """soustraction de deux de vecteurs"""
        return Point(self.x - p.x, self.y - p.y)

    def angle(self):
        """retourne l'angle du vecteur"""
        return math.atan2(self.y, self.x)

    def __eq__(self, a) -> bool:
        """retourne True si les deux points ``self`` et ``a`` sont egaux,
        False sinon"""
        return self.x == a.x and self.y == a.y


class Segment:
    """
    Définit un segment, soit deux @see cl Point.
    """

    # voir le commentaire associees a la ligne contenant __slots__
    # dans la classe Point
    __slots__ = "a", "b"

    def __init__(self, a, b):
        """
        constructeur, pour éviter des erreurs d'etourderie,
        on crée des copies des extrémités a et b,
        comme ce sont des classes, une simple affectation ne suffit pas
        """
        self.a, self.b = copy.copy(a), copy.copy(b)

    def __str__(self) -> str:
        """permet d'afficher le segment avec l'instruction print"""
        return "[{0},{1}]".format(self.a, self.b)

    def directeur(self) -> Point:
        """retourne le vecteur directeur du segment,
        ce vecteur est norme"""
        p = Point(self.b.x - self.a.x, self.b.y - self.a.y)
        p.normalise()
        return p

    def normal(self) -> float:
        """retourne le vecteur normal du segment,
        ce vecteur est norme"""
        p = Point(self.a.y - self.b.y, self.b.x - self.a.x)
        p.normalise()
        return p

    def first(self):
        """Retourne la première extrémité."""
        return self.a

    def last(self):
        """Retourne la seconde extrémité."""
        return self.b



"""
@file
@brief Functions to retrieve data from Wikipedia
"""
import os
from pyquickhelper.loghelper import noLOG
from pyquickhelper.filehelper import get_url_content_timeout, ungzip_files
from .data_exceptions import DataException


def download_pageviews(dt, folder=".", unzip=True, timeout=-1,
                       overwrite=False, fLOG=noLOG):
    """
    Downloads wikipedia pagacount for a precise date (up to the hours),
    the url follows the pattern::

        https://dumps.wikimedia.org/other/pageviews/%Y/%Y-%m/pagecounts-%Y%m%d-%H0000.gz

    @param      dt          datetime
    @param      folder      where to download
    @param      unzip       unzip the file
    @param      timeout     timeout
    @param      overwrite   overwrite
    @param      fLOG        logging function
    @return                 filename

    More information on page
    `pageviews <https://dumps.wikimedia.org/other/pageviews/>`_.
    """
    url = "https://dumps.wikimedia.org/other/pageviews/%Y/%Y-%m/pageviews-%Y%m%d-%H0000.gz"
    url = dt.strftime(url)
    file = url.split("/")[-1]
    name = os.path.join(folder, file)
    unzipname = os.path.splitext(name)[0]
    if overwrite or (not os.path.exists(name) and not os.path.exists(unzipname)):
        get_url_content_timeout(url, timeout=timeout,
                                encoding=None, output=name, chunk=2**20, fLOG=fLOG)
    if unzip and not os.path.exists(unzipname):
        names = ungzip_files(name, unzip=False, where_to=folder)
        os.remove(name)
        if isinstance(names, list):
            if len(names) != 1:
                raise DataException(  # pragma: no cover
                    "Expecting only one file, not '{0}'".format(names))
            return names[0]
        return names
    return name


def download_dump(country, name, folder=".", unzip=True, timeout=-1,
                  overwrite=False, fLOG=noLOG):
    """
    Downloads *wikipedia dumps* from
    `dumps.wikimedia.org/frwiki/latest/
    <https://dumps.wikimedia.org/frwiki/latest/>`_.

    @param      country     country
    @param      name        name of the stream to download
    @param      folder      where to download
    @param      unzip       unzip the file
    @param      timeout     timeout
    @param      overwrite   overwrite
    @param      fLOG        logging function
    """
    url = "https://dumps.wikimedia.org/{0}wiki/latest/{0}wiki-{1}".format(
        country, name)
    file = url.split("/")[-1]  # pylint: disable=C0207
    name = os.path.join(folder, file)
    unzipname = os.path.splitext(name)[0]
    if overwrite or (not os.path.exists(name) and not os.path.exists(unzipname)):
        get_url_content_timeout(url, timeout=timeout,
                                encoding=None, output=name, chunk=2**20, fLOG=fLOG)
    if unzip and not os.path.exists(unzipname):
        names = ungzip_files(name, unzip=False, where_to=folder)
        os.remove(name)
        if isinstance(names, list):
            if len(names) != 1:
                raise DataException(  # pragma: no cover
                    "Expecting only one file, not '{0}'".format(names))
            return names[0]
        return names
    return name[:-3] if name.endswith('.gz') else name


def download_titles(country, folder=".", unzip=True, timeout=-1,
                    overwrite=False, fLOG=noLOG):
    """
    Downloads wikipedia titles from
    `dumps.wikimedia.org/frwiki/latest/latest-all-titles-in-ns0.gz
    <https://dumps.wikimedia.org/frwiki/latest/latest-all-titles-in-ns0.gz>`_.

    @param      country     country
    @param      folder      where to download
    @param      unzip       unzip the file
    @param      timeout     timeout
    @param      overwrite   overwrite
    @param      fLOG        logging function
    """
    return download_dump(country, "latest-all-titles-in-ns0.gz",
                         folder, unzip=unzip, timeout=timeout,
                         overwrite=overwrite, fLOG=fLOG)


def normalize_wiki_text(text):
    """
    Normalizes a text such as a wikipedia title.

    @param      text        text to normalize
    @return                 normalized text
    """
    return text.replace("_", " ").replace("''", '"')


def enumerate_titles(filename, norm=True, encoding="utf8"):
    """
    Enumerates titles from a file.

    @param      filename        filename
    @param      norm            normalize in the function
    @param      encoding        encoding
    """
    if norm:
        with open(filename, "r", encoding=encoding) as f:
            for line in f:
                yield normalize_wiki_text(line.strip(" \r\n\t"))
    else:
        with open(filename, "r", encoding=encoding) as f:
            for line in f:
                yield line.strip(" \r\n\t")
